{
  Name    :   ClientUtils.pas

  Purpose :   Provides services for the client accessing Delphi Bank components
               by managing the lifetime of, and providing easy access to these
               components.

  Usage   :   This source code is supplied as example code with Delphi Informant
               magazine.
              The author makes no guarantee for its correctness or fitness
               for any purpose and cannot be held responsible for any loss
               or damage that may occur through its use or misuse.
              You may copy, modify or use any part of this source code
               provided the author is credited for his work.

  Author  :   Paul M. Fairhurst, September 1998. paul@c-s-c.demon.co.uk
              (Based on code kindly donated by Mark Smith)
}
unit ClientUtils;

interface

uses
  Grids,
  Windows,
  DelphiBankServer2_TLB,
  AccountTypes;

type
  TClientServices = class
  private
    FLocalAccountTypes : TAccountTypeCollection;
    FAccountsTypes: IAccountTypes2;
    FClientKey: integer;
    FCustomers: ICustomers;
    FAccounts: IAccounts;
    FTransactions: ITransactions;
    private
      procedure SetAccountsTypes(const Value: IAccountTypes2);
      procedure SetClientKey(const Value: integer);
      procedure SetCustomers(const Value: ICustomers);
      procedure SetAccounts(const Value: IAccounts);
      procedure SetTransactions(const Value: ITransactions);

      // Get functions to allow 'lazy' instantiation
      function GetAccounts: IAccounts;
      function GetAccountsTypes: IAccountTypes2;
      function GetCustomers: ICustomers;
      function GetTransactions: ITransactions;
      function GetLocalAccountTypes: TAccountTypeCollection;
    public
      class procedure ShutDown;
      function GetAccountType(index: integer): TAccountType;
      // Identifies this session
      property ClientKey : integer read FClientKey write SetClientKey;

      // Instances of the interfaces
      Property AccountsTypes: IAccountTypes2 read GetAccountsTypes write SetAccountsTypes;
      Property Customers : ICustomers read GetCustomers write SetCustomers;
      property Accounts  : IAccounts read GetAccounts write SetAccounts;
      property Transactions : ITransactions read GetTransactions write SetTransactions;

      // Used to make it easier to get at account type details
      property LocalAccountTypes : TAccountTypeCollection read GetLocalAccountTypes;
      property AccountType [index : integer] : TAccountType read GetAccountType;
  end;


var
  ClientServices : TClientServices;

procedure StringGridFromResultSet(AStringGrid : TStringGrid;
                                       const varResultSet: OleVariant);

implementation

uses Dialogs;

procedure StringGridFromResultSet(AStringGrid : TStringGrid;
                                       const varResultSet: OleVariant);
var
  nColLow, nColHigh, nRowLow, nRowHigh : Integer;
  x, y : integer;
begin
  if VarArrayDimCount(varResultSet) <> 2 then begin
    ShowMessage('ShowResultSetInStringGrid() - Expecting a 2 dimensional array as result set.');
    Exit;
  end;

  { Work out the size of the array. }
  nColLow  := VarArrayLowBound(varResultSet, 1);
  nColHigh := VarArrayHighBound(varResultSet, 1);
  nRowLow  := VarArrayLowBound(varResultSet, 2);
  nRowHigh := VarArrayHighBound(varResultSet, 2);

  { Set up the grid. }
  with AStringGrid do begin
    FixedCols := 0;
    FixedRows := 1;
    ColCount := nColHigh-nColLow;
    RowCount := nRowHigh-nRowLow;
    for y := 0 to RowCount do
      for x := 0 to ColCount do
        Cells[x,y] := varResultSet[x,y];
  end;
end;


{ TClientServices }

function TClientServices.GetAccounts: IAccounts;
begin
  if FAccounts = nil then
    FAccounts := CoAccounts.Create;

  result := FAccounts;
end;

function TClientServices.GetAccountsTypes: IAccountTypes2;
begin
  if FAccountsTypes = nil then
    FAccountsTypes := CoAccountTypes2.Create;

  Result := FAccountsTypes;
end;


function TClientServices.GetCustomers: ICustomers;
begin
  if FCustomers = nil then
    FCustomers := CoCustomers.Create;

  result := FCustomers;
end;


function TClientServices.GetTransactions: ITransactions;
begin
  if FTransactions = nil then
    FTransactions := CoTransactions.Create;

  Result := FTransactions;
end;


function TClientServices.GetAccountType(index: integer): TAccountType;
var
  i : integer;
  AccountType : TAccountType;
begin
  Assert (LocalAccountTypes <> nil);

  Result := nil;

  for i := 0 to LocalAccountTypes.Count-1 do
  begin
    AccountType := LocalAccountTypes.Items[i] as TAccountType;
    if AccountType.AccountTypeID = index then
    begin
      Result := AccountType;
      exit;
    end;
  end;

end;

function TClientServices.GetLocalAccountTypes: TAccountTypeCollection;
begin
  if FLocalAccountTypes = nil then
  begin
    FLocalAccountTypes := TAccountTypeCollection.Create (TAccountType);
    FLocalAccountTypes.Load;
  end;
  Result := FLocalAccountTypes;
end;


procedure TClientServices.SetAccounts(const Value: IAccounts);
begin
  FAccounts := Value;
end;

procedure TClientServices.SetAccountsTypes(const Value: IAccountTypes2);
begin
  FAccountsTypes := Value;
end;

procedure TClientServices.SetClientKey(const Value: integer);
begin
  FClientKey := Value;
end;

procedure TClientServices.SetCustomers(const Value: ICustomers);
begin
  FCustomers := Value;
end;


procedure TClientServices.SetTransactions(const Value: ITransactions);
begin
  FTransactions := Value;
end;

class procedure TClientServices.ShutDown;
begin
  ClientServices.AccountsTypes := Nil;
  ClientServices.Customers := Nil;
  ClientServices.Accounts  := Nil;
  ClientServices.Transactions := Nil;
  ClientServices.Free;
  ClientServices := nil;
end;

initialization
    ClientServices := TClientServices.Create;

finalization
  ClientServices.Free;
  ClientServices := nil;


end.
