{***************************************************************}
{                                                               }
{  Formula Compiler Library for Delphi ( 16 & 32-bit )          }
{                                                               }
{  Copyright (c) 1996, 1997  Dimak & Vlad.S                     }
{                                                               }
{  e-mail:  dimak@uniyar.ac.ru  &  vlads@uniyar.ac.ru           }
{                                                               }
{***************************************************************}

unit FCLib;

{$IFDEF WIN32}

{$S-,H+,B-}

{$ELSE}

{$S-,F+,O-,B-}

{$ENDIF}

interface

uses SysUtils, Classes, Consts, Forms;

{ TFormulaLib & TFunctionList }
type
  TFormulaLib = class;

  { TFunctionList - list to hold aliases and user functions }
  TFunctionList = class(TPersistent)
  private
    FList: TStrings;
    FParent: TFormulaLib;
    function GetArgCount(Index: Integer): Integer;
    function GetCount: Integer;
    function GetNames(Index: Integer): String;
    function GetSources(Index: Integer): String;
    procedure SetNames(Index: Integer; NewValue: String);
    procedure SetSources(Index: Integer; NewValue: String);
    function IsUnusedIdent(const S: String): Boolean;
    procedure Assign(Source: TPersistent);
    procedure ReadData(Reader: TReader);
    procedure WriteData(Writer: TWriter);
  public
    constructor Create(AParent: TFormulaLib);
    destructor Destroy; override;
    { add entry with name AName and source ASource to the list }
    function Add(AName, ASource: String): Integer;
    { delete entry Index }
    procedure Delete(Index: Integer);
    { returns number of entry with name AName }
    function IndexOf(AName: String): Integer;
    { number of arguments }
    property ArgCount[Index: Integer]: Integer read GetArgCount;
    { number of elements in the list }
    property Count: Integer read GetCount;
    { names of entries }
    property Names[Index: Integer]: String read GetNames write SetNames;
    { sources of entries }
    property Sources[Index: Integer]: String read GetSources write SetSources;
  end;

  { TFormulaLib - holds user aliases and user funtions }
  TFormulaLib = class(TComponent)
  private
    FAliases, FFunctions: TFunctionList;
    procedure SetAliases(NewValue: TFunctionList);
    procedure SetFunctions(NewValue: TFunctionList);
    function CanDeleteName(Name: String): Boolean;
  protected
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent);
    { edit library contents - show TFormulaLibMan dialog }
    function Edit(APage: Integer): Boolean;
    { expand aliases - replace aliases in S by their source }
    function ExpandAliases(const S: String): String;
    { check validity  }
    function ValidFunctionName(const S: String): Integer;
    function ValidAliasName(const S: String): Integer;
    { list of aliases }
    property Aliases: TFunctionList read FAliases write SetAliases;
    { list of functions }
    property Functions: TFunctionList read FFunctions write SetFunctions;
  end;

implementation

uses WinTypes, WinProcs, FCUtils, FCLibMan;

{ TFormulaLib methods }
constructor TFormulaLib.Create;
begin
  inherited Create(AOwner);
  FAliases := TFunctionList.Create(Self);
  FFunctions := TFunctionList.Create(Self);
end;

destructor TFormulaLib.Destroy;
begin
  FAliases.Free;
  FFunctions.Free;
  inherited Destroy;
end;

procedure TFormulaLib.SetAliases(NewValue: TFunctionList);
begin
  FAliases.Assign(NewValue);
end;

procedure TFormulaLib.SetFunctions(NewValue: TFunctionList);
begin
  FFunctions.Assign(NewValue);
end;

procedure TFormulaLib.Assign(Source: TPersistent);
var
  SourceName: String;
begin
  if Source <> nil
    then SourceName := Source.ClassName
    else SourceName := 'nil';
  if SourceName <> ClassName
{$IFDEF VER100}
    then raise EConvertError.CreateFmt(SAssignError, [SourceName, ClassName])
{$ELSE}
    then raise EConvertError.CreateResFmt(SAssignError, [SourceName, ClassName])
{$ENDIF}
    else begin
      FAliases.Assign((Source as TFormulaLib).FAliases);
      FFunctions.Assign((Source as TFormulaLib).FFunctions);
    end;
end;

function TFormulaLib.ExpandAliases(const S: String): String;
var
  Index, j: Integer;
begin
  Result := S;
  for j := 0 to Aliases.Count - 1 do
    repeat
      Index := ContainsIdent(Result, Aliases.Names[j]);
      if Index > 0
        then begin
          Delete(Result, Index, Length(Aliases.Names[j]));
          Insert(ExpandAliases(Aliases.Sources[j]), Result, Index);
        end;
    until Index < 0;
end;

{ Checks the validity of function or constant name given in S }
function TFormulaLib.ValidFunctionName(const S: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to FFunctions.Count - 1 do
    if CompareText(S, FFunctions.Names[i]) = 0
      then begin
        Result := i;
        break;
      end;
end;

{ Checks the validity of alias name given in S }
function TFormulaLib.ValidAliasName(const S: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to FAliases.Count - 1 do
    if CompareText(S, FAliases.Names[i]) = 0
    then begin
      Result := i;
      break;
    end;
end;

{ Returns TRUE if entry Name don't referenced by other entries  }
function TFormulaLib.CanDeleteName(Name: String): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Aliases.Count - 1 do
    if ContainsIdent(Aliases.Sources[i], Name) > 0
      then exit;
  for i := 0 to Functions.Count - 1 do
    if ContainsIdent(Functions.Sources[i], Name) > 0
      then exit;
  Result := True;
end;

procedure TFormulaLib.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('Functions', FFunctions.ReadData, FFunctions.WriteData, True);
  Filer.DefineProperty('Aliases', FAliases.ReadData, FAliases.WriteData, True);
end;

{ Shows TFormulaLibMan, allows user edit library contents }
function TFormulaLib.Edit(APage: Integer): Boolean;
var
  FLM: TFormulaLibMan;
begin
  FLM := TFormulaLibMan.Create(Application);
  try
    Result := FLM.Edit(Self, APage);
  finally
    FLM.Free;
  end;
end;

{ TFunctionList methods }
constructor TFunctionList.Create;
begin
  inherited Create;
  if (AParent = nil) or not (AParent is TFormulaLib)
    then raise EFCError.Create('');
  FParent := AParent;
  FList := TStringList.Create;
end;

destructor TFunctionList.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

{ Checks if Identifier given in S is unused }
function TFunctionList.IsUnusedIdent(const S: String): Boolean;
begin
  Result := (ValidBIFName(S) < 0) and (FParent.ValidFunctionName(S) < 0) and
            (FParent.ValidAliasName(S) < 0)
end;

function TFunctionList.Add(AName, ASource: String): Integer;
var
  InternalExpr: TInternalExprArray;
begin
{$IFDEF DEMO}
  if Count > 3
    then begin
  {$IFDEF WIN32}
      if IsConsole
      then WriteLn(#10#13 + sLimit)
      else
  {$ENDIF}
      MessageBox(0, sLimit, sInfo, mb_Ok or mb_IconInformation);
      exit;
    end;
{$ENDIF}
  if IsValidIdent(AName)
  then if IsUnusedIdent(AName)
       then begin
         InternalExpr := TInternalExprArray.Create;
         Result := MakeInternalExpr(ASource, InternalExpr, nil, FParent);
         InternalExpr.Free;
         if Result = fcOk
         then begin
           FList.Add(AName);
           FList.Add(ASource);
         end
       end
       else Result := fcReservedId
   else Result := fcInvalidId;
end;

procedure TFunctionList.Delete(Index: Integer);
begin
  if FParent.CanDeleteName(FList[2*Index])
  then begin
    FList.Delete(2*Index);
    FList.Delete(2*Index);
  end
  else FCCheck(fcUsedName);
end;

function TFunctionList.IndexOf(AName: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to FList.Count div 2 - 1 do
    if CompareText(FList[2*i], AName) = 0
      then begin
        Result := i;
        break;
      end;
end;

function TFunctionList.GetCount: Integer;
begin
  Result := FList.Count div 2;
end;

function TFunctionList.GetArgCount(Index: Integer): Integer;
begin
  Result := Arguments(FList[2*Index + 1], FParent)
end;

procedure TFunctionList.Assign(Source: TPersistent);
var
  SourceName: String;
begin
  if Source <> nil
    then SourceName := Source.ClassName
    else SourceName := 'nil';
  if SourceName <> ClassName
{$IFDEF VER100}
    then raise EConvertError.CreateFmt(SAssignError, [SourceName, ClassName])
{$ELSE}
    then raise EConvertError.CreateResFmt(SAssignError, [SourceName, ClassName])
{$ENDIF}
    else FList.Assign((Source as TFunctionList).FList);
end;

function TFunctionList.GetNames(Index: Integer): String;
begin
  Result := FList[2*Index];
end;

function TFunctionList.GetSources(Index: Integer): String;
begin
  Result := FList[2*Index + 1];
end;

procedure TFunctionList.SetNames(Index: Integer; NewValue: String);
begin
  if CompareText(FList[2*Index], NewValue) <> 0
  then if IsValidIdent(NewValue)
         then if IsUnusedIdent(NewValue)
                then if FParent.CanDeleteName(FList[2*Index])
                       then FList[2*Index] := NewValue
                       else FCCheck(fcUsedName)
                else FCCheck(fcReservedId)
         else FCCheck(fcInvalidId)
end;

procedure TFunctionList.SetSources(Index: Integer; NewValue: String);
var
  Res: Integer;
  InternalExpr: TInternalExprArray;
begin
  if CompareText(FList[2*Index + 1], NewValue) <> 0
  then begin
    InternalExpr := TInternalExprArray.Create;
    Res := MakeInternalExpr(NewValue, InternalExpr, nil, FParent);
    InternalExpr.Free;
    if Res = fcOk
      then FList[2*Index + 1] := NewValue
      else FCCheck(Res)
  end;
end;

procedure TFunctionList.WriteData(Writer: TWriter);
var
  I: Integer;
begin
  Writer.WriteListBegin;
  for I := 0 to FList.Count - 1 do Writer.WriteString(FList[I]);
  Writer.WriteListEnd;
end;

procedure TFunctionList.ReadData(Reader: TReader);
begin
  Reader.ReadListBegin;
  FList.Clear;
  while not Reader.EndOfList do FList.Add(Reader.ReadString);
  Reader.ReadListEnd;
end;

end.
