{***************************************************************}
{                                                               }
{  Formula Compiler Pack for Delphi ( 16 & 32-bit ) ver 3.0     }
{                                                               }
{  Copyright (c) 1996, 1997  Dimak & Vlad.S                     }
{                                                               }
{  e-mail:  dimak@uniyar.ac.ru  &  vlads@uniyar.ac.ru           }
{                                                               }
{***************************************************************}

unit FCUtils;

interface

uses SysUtils, Classes, Consts, FCLib;

{ Exception }
type
  EFCError = class(Exception);

{ Maximum Array size }
const
  MaxArraySize = MaxInt - 1;

{ Error codes }
const
  fcOk = 0;
  fcInternal = -1;
  fcInvalidId = fcInternal - 1;
  fcReservedId = fcInternal - 2;
  fcUsedName = fcInternal - 3;
  fcUnknownVar = fcInternal - 4;
  fcUnknownFunc = fcInternal - 5;
const
  aArray = fcInternal - 10;
  aIndex = aArray - 1;
  aTypecast = aArray - 2;
  aCapacity = aArray - 3;

{ Error handler }
const
  FCCheck: procedure(ErrorCode: Integer) = nil;

{ Check if a Source contains identifier Ident }
function ContainsIdent(const Source, Ident: String): Integer;

{ Getting arguments count }
function Arguments(const S: String; Lib: TFormulaLib): Integer;

{ Checks the validity of logic operator given in S }
function ValidLogicOpName(const S: String): Integer;

{ Checks the validity of function or constant name given in S }
function ValidBIFName(const S: String): Integer;

{ TArray class }
{ array of items which size is FItemSize  }
{ TArray holds items in on chunk of memory, can be }
{ used to hold record, integers, ... }
type
  TArray = class(TObject)
  protected
    FArray: PChar;
    FCount: Integer;
    FCapacity: Integer;
    FItemSize: Integer;
    procedure Error(ErrorCode: Integer); virtual;
    procedure Grow; virtual;
    procedure SetCapacity(NewCapacity: Integer);
  public
    constructor Create(AItemSize: Integer);
    destructor Destroy; override;
    function Add(const Item): Integer;
    procedure Assign(AArray: TArray);
    procedure Clear;
    procedure GetItem(Index: Integer; var Item);
    procedure Delete(Index: Integer);
    function IndexOf(const Item): Integer;
    procedure Insert(Index: Integer; const Item);
    procedure SetItem(Index: Integer; const Item);
    property Data: PChar read FArray;
    property Count: Integer read FCount;
  end;

type
  { TInternalItem holds different elements of expression }
  TInternalExprItem = record
    Elem: Char;  { type of element :  'f': built-in function }
                 {                    'c': constant          }
                 {                    'a': argument          }
    ID: Integer; {  identifier :  built-in function number, ...}
    Cnst: Extended;   {  value of constants  }
  end;

  { TInternalExprArray  - internal representation of expression to be compiled }
  TInternalExprArray = class(TArray)
  private
    function GetElem(Index: Integer): Char;
    function GetId(Index: Integer): Integer;
    function GetCnst(Index: Integer): Extended;
    function Get(Index: Integer): TInternalExprItem;
    procedure SetElem(Index: Integer; NewValue: Char);
    procedure SetId(Index: Integer; NewValue: Integer);
    procedure SetCnst(Index: Integer; NewValue: Extended);
  public
    constructor Create;
    procedure Add(Item: TInternalExprItem);
    procedure Insert(Index: Integer; Item: TInternalExprItem);
    property Elem[Index: Integer]: Char read GetElem write SetElem;
    property Id[Index: Integer]: Integer read GetId write SetId;
    property Cnst[Index: Integer]: Extended read GetCnst write SetCnst;
    property Items[Index: Integer]: TInternalExprItem read Get;
  end;

{ Makes TInternalExprItem }
function MakeItem(AElem: Char; AID: Integer; ACnst: Extended): TInternalExprItem;

{ Makes InternalExpr }
{ S - expression to compile, InternalExpr - internal representation  }
{ NameList - list of unknown names (funtions which are not defined in the }
{ library Lib }
function MakeInternalExpr(S: String; InternalExpr: TInternalExprArray; NameList: TStringList; Lib: TFormulaLib): Integer;

{ Makes RPN from InternalExpr }
{ Produces Reverse Polish Notation of InternalExpr and put it into Postfix }
procedure MakePostfix(InternalExpr, Postfix: TInternalExprArray);

{ Optimize InternalExpr }
{ Performs some optimization procedures : convert unary '-' ... }
procedure Optimize(InternalExpr: TInternalExprArray);

{ TCodeArray  - this class is used to hold machine code }
type
  TCodeArray = class(TArray)
  public
    constructor Create;
    procedure Add(AB: Byte);                      { add byte to array    }
    procedure AddInteger(AL: Integer);            { add integer to array }
    procedure AddCode(AC: PChar);
    function  GetInteger(Index: Integer): Integer;
    procedure InsertInteger(Index: Integer; AL: Integer);
    procedure InsertCode(Index: Integer; AC: PChar);
  end;

{ TIntArray  -- array of integer }
  TIntArray = class(TArray)
  private
    function GetInt(Index: Integer): Integer;
    procedure SetInt(Index: Integer; NewValue: Integer);
  public
    constructor Create;
    procedure Add(AL: Integer);
    property Item[Index: Integer]: Integer read GetInt write SetInt;
  end;

{ User functions and constants must have Index >= UserFuncBase }
const
  UserFuncBase = $100;

{ Settings for analyzer (parser) }
const
  Numbers   : set of char = ['0'..'9'];
  Operations: set of char = ['+', '-', '*', '/', '^', '!'];
  Letters   : set of char = ['A'..'Z','_', 'a'..'z', '$'];
  Operators : set of char = ['<', '>', '='];
  ArgumentMarker = '%';
  ArgumentSeparator = ',';
  DecimalSeparator = '.';

{ Built-in logic operators }
type
  TLogicOpName = String[2]; { examples  <>, <=, ... }
  TLogicOpCode = PChar;     { machine code }

  TLogicOp = record
    Name: TLogicOpName;
    Code: TLogicOpCode;
  end;

{ array of available precedence operators and their machine code  }
{ Structure of Code: }
{   First byte:   Code length  }
{   ....      :   FPU  code    }
const
  LogicOpCount = 6;
  BuiltInOperators : array [0..LogicOpCount - 1] of TLogicOp = (
{$IFDEF WIN32}
  (Name: '>'; Code: #$10#$50#$DE#$D9#$9B#$DF#$E0#$9E#$77#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8#$58),
  { PUSH ax (50); FCOMPP (DED9); FWAIT (9B); FNSTSW AX (DFE0);  SAHF (9E); }
  { JA (7704); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8); POP ax (58) }
  (Name: '<'; Code: #$10#$50#$DE#$D9#$DF#$E0#$9B#$9E#$72#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8#$58),
  { PUSH ax (50); FCOMPP (DED9); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E); }
  { JB (7204); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8); POP ax (58) }
  (Name: '>='; Code: #$10#$50#$DE#$D9#$DF#$E0#$9B#$9E#$73#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8#$58),
  { PUSH ax (50); FCOMPP (DED9); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E); }
  { JAE (7304); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8); POP ax (58) }
  (Name: '<='; Code: #$10#$50#$DE#$D9#$DF#$E0#$9B#$9E#$76#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8#$58),
  { PUSH ax (50); FCOMPP (DED9); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E); }
  { JBE (7604); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8); POP ax (58) }
  (Name: '='; Code: #$10#$50#$DE#$D9#$9B#$DF#$E0#$9E#$74#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8#$58),
  { PUSH ax (50); FCOMPP (DED9); FWAIT (9B); FNSTSW AX (DFE0);  SAHF (9E); }
  { JE (7404); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8); POP ax (58) }
  (Name: '<>'; Code: #$10#$50#$DE#$D9#$DF#$E0#$9B#$9E#$75#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8#$58)
  { PUSH ax (50); FCOMPP (DED9); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E); }
  { JNE (7504); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8); POP ax (58) }
{$ELSE}
  (Name: '>'; Code: #$0E#$DE#$D9#$9B#$DF#$E0#$9E#$77#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8),
  { FCOMPP (DED9); FWAIT (9B); FNSTSW AX (DFE0);  SAHF (9E); }
  { JA (7704); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8) }
  (Name: '<'; Code: #$0E#$DE#$D9#$DF#$E0#$9B#$9E#$72#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8),
  { FCOMPP (DED9); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E); }
  { JB (7204); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8) }
  (Name: '>='; Code: #$0E#$DE#$D9#$DF#$E0#$9B#$9E#$73#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8),
  { FCOMPP (DED9); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E); }
  { JAE (7304); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8) }
  (Name: '<='; Code: #$0E#$DE#$D9#$DF#$E0#$9B#$9E#$76#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8),
  { FCOMPP (DED9); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E); }
  { JBE (7604); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8) }
  (Name: '='; Code: #$0E#$DE#$D9#$9B#$DF#$E0#$9E#$74#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8),
  { FCOMPP (DED9); FWAIT (9B); FNSTSW AX (DFE0);  SAHF (9E); }
  { JE (7404); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8) }
  (Name: '<>'; Code: #$0E#$DE#$D9#$DF#$E0#$9B#$9E#$75#$04#$D9+
                    #$EE#$EB#$02#$D9#$E8)
  { FCOMPP (DED9); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E); }
  { JNE (7504); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8) }
{$ENDIF}
  );

{ Built in Functions and Constants }
type
  TBIFName = String[7];  { functon name  }
  TBIFCode = PChar;      { function code }
  TBIFArgs = -1..3;      { number of arguments }

  TBIFD = record
    Name: TBIFName;
    Args: TBIFArgs;
    Code: TBIFCode;
  end;

{ array of available built-in functions and their machine code  }
{ Structure of Code: }
{   First byte:   Code length  }
{   ....      :   FPU  code    }
{ Note: 16 and 32 bit code is different in some cases }
const
  BIFCount = 38;
  BuiltInFuncs: array[0..BIFCount-1] of TBIFD = (
{*}(Name:'CHS';  Args: 1; Code: #$02#$D9#$E0),
   { FCHS (D9E0) }
{*}(Name:'ZERO'; Args: 0; Code: #$02#$D9#$EE),
   { FLDZ (D9EE) }
{*}(Name:'ONE';  Args: 0; Code: #$02#$D9#$E8),
   { FLD1 (D9E8) }
{*}(Name:'PI';  Args: 0;  Code: #$02#$D9#$EB),
   { FLDPI (D9EB) }
{*}(Name:'LN2'; Args: 0; Code: #$02#$D9#$ED),
   { FLDLN2 (D9ED) }
{$IFDEF WIN32}
{*}(Name:'NOT'; Args: 1; Code: #$12#$50#$D9#$E4#$DF#$E0#$DD#$D8#$9B#$9E#$74#$04 +
                                   #$D9#$EE#$EB#$02#$D9#$E8#$58),
   { PUSH ax (50); FTST (D9E4); FNSTSW ax (DFE0); FSTP st(0) (DDD8); FWAIT (9B);
   { SAHF (9E); JE (7404); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8); POP ax (58)   }
{$ELSE}
{*}(Name:'NOT'; Args: 1; Code: #$10#$D9#$E4#$DF#$E0#$DD#$D8#$9B#$9E#$74#$04 +
                                   #$D9#$EE#$EB#$02#$D9#$E8),
   { FTST (D9E4); FNSTSW ax (DFE0); FSTP st(0) (DDD8); FWAIT (9B); SAHF (9E); }
   { JE (7404); FLDZ (D9EE); JMP (EB02), FLD1 (D9E8);                         }
{$ENDIF}
{*}(Name:'POW'; Args: 2; Code: #$14#$D9#$F1#$D9#$C0#$D9#$FC +
                                   #$D9#$C9#$D8#$E1#$D9#$F0#$D9#$E8 +
                                   #$DE#$C1#$D9#$FD#$DD#$D9),
   { FYL2X (D9F1); FLD ST(0) (D9C0); FRNDINT (D9FC) }
   { FXCH (D9C9); FSUB ST,ST(1) (D8E1); F2XM1 (D9F0); FLD1 (D9E8);     }
   { FADDP ST(1),ST(0) (DEC1); FSCALE (D9FD); FSTP ST(1) (DDD9);       }
{ }(Name:'SIN';  Args: 1; Code: #$02#$D9#$FE),
   { FSIN (D9FE) }
{ }(Name:'COS';  Args: 1; Code: #$02#$D9#$FF),
   { FCOS (D9FF) }
{ }(Name:'TAN';  Args: 1; Code: #$04#$D9#$F2#$DD#$D8),
   { FPTAN (D9F2); FSTP ST(0) (DDD8) }
{ }(Name:'COTAN';  Args: 1; Code: #$04#$D9#$F2#$DE#$F1),
   { FPTAN (D9F2); FDIVRP (DEF1) }
{ }(Name:'ATAN'; Args: 1; Code: #$04#$D9#$E8#$D9#$F3),
   { FLD1 (D9E8); FPATAN (D9F3) }
{ }(Name:'ACOS'; Args: 1; Code: #$0E#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1 +
                                    #$D9#$FA#$D9#$C9#$D9#$F3),
   { FLD ST(0)  (D9C0);  FMUL ST(0), ST (D8C8);  FLD1 (D9E8);           }
   { FSUBRP ST(1), ST (DEE1); FSQRT (D9FA); FXCH (D9C9); FPATAN (D9F3); }
{ }(Name:'ASIN'; Args: 1; Code: #$0C#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1 +
                                    #$D9#$FA#$D9#$F3),
   { FLD ST(0)  (D9C0);  FMUL ST(0), ST (D8C8);  FLD1 (D9E8);           }
   { FSUBRP ST(1), ST (DEE1); FSQRT (D9FA); FPATAN (D9F3);              }
{ }(Name:'ABS';  Args: 1; Code: #$02#$D9#$E1),
{ }(Name:'SQRT'; Args: 1; Code: #$02#$D9#$FA),
{ }(Name:'EXP';  Args: 1; Code: #$16#$D9#$EA#$DE#$C9#$D9#$C0#$D9#$FC +
                                    #$D9#$C9#$D8#$E1#$D9#$F0#$D9#$E8 +
                                    #$DE#$C1#$D9#$FD#$DD#$D9),
   { FLDL2E (D9EA); FMULP ST(1),ST(0) (DEC9); FLD ST(0) (D9C0);        }
   { FRNDINT (D9FC); FXCH (D9C9); FSUB ST,ST(1) (D8E1)                 }
   { F2XM1 (D9F0); FLD1 (D9E8); FADDP ST(1),ST(0) (DEC1);              }
   { FSCALE (D9FD); FSTP ST(1) (DDD9);                                 }
{ }(Name:'LN'; Args: 1;Code: #$06#$D9#$ED#$D9#$C9#$D9#$F1),
   { FLDLN2 (D9ED); FXCH ST(1) (D9C9); FYL2X (D9F1) }
{ }(Name:'LG'; Args: 1;Code: #$06#$D9#$EC#$D9#$C9#$D9#$F1),
   { FLDLG2 (D9ED); FXCH ST(1) (D9C9); FYL2X (D9F1) }
{$IFDEF WIN32}
{ }(Name:'INT';Args: 1; Code: #$17#$D9#$7C#$24#$FC#$9B#$66#$C7#$44#$24 +
                                  #$F8#$32#$1F#$D9#$6C#$24#$F8#$D9#$FC +
                                  #$9B#$D9#$6C#$24#$FC),
   { FSTCW mem0 (D97C24FC); FWAIT (9B);                                   }
   { MOV mem2, $1F32 (66C74424F8321F); FLDCW mem2 (D96C24F8);             }
   { FRNDINT (D9FC); FWAIT (9B); FLDCW mem0 (D96C24FC)                    }
{ }(Name:'FRAC';Args: 1; Code: #$1B#$D9#$C0#$D9#$7C#$24#$FC#$9B +
                                   #$66#$C7#$44#$24#$F8#$32#$1F +
                                   #$D9#$6C#$24#$F8#$D9#$FC#$9B +
                                   #$D9#$6C#$24#$FC#$DE#$E9),
   { FLD ST(0) (D9C0); FSTCW mem0 (D97C24FC); FWAIT (9B); }
   { MOV mem2, $1F32 (66C74424F8321F);                    }
   { FLDCW mem2 (D96C24F8); FRNDINT (D9FC); FWAIT (9B);   }
   { FLDCW mem0 (D96C24FC) FSUBP st(0),st(1) (DEE9);      }
{$ELSE}
{ }(Name:'INT';Args: 1; Code: #$15#$D9#$7E#$F4#$9B#$8B#$46#$F4 +
                                  #$0D#$00#$0C#$89#$46#$F2 +
                                  #$D9#$6E#$F2#$D9#$FC#$D9#$6E#$F4),
   { FSTCW mem0 (D97DF4); FWAIT (9B); MOV ax,mem0 (8B45F4)      }
   { OR ax,0C00 (0D000C); MOV mem2,ax (8945F2);                 }
   { FLDCW mem2 (D96DF2); FRNDINT (D9FC); FLDCW mem0 (D96DF4)   }
{ }(Name:'FRAC';Args: 1; Code: #$19#$D9#$7E#$F4#$9B#$8B#$46#$F4 +
                                   #$0D#$00#$0C#$89#$46#$F2 +
                                   #$D9#$6E#$F2#$D9#$C0#$D9#$FC +
                                   #$DE#$E9#$D9#$6E#$F4),
   { FSTCW mem0 (D97EF4); FWAIT (9B); MOV ax,mem0 (8B46F4)      }
   { OR ax,0C00 (0D000C); MOV mem2,ax (8946F2);                 }
   { FLDCW mem2 (D96EF2); FLD st(0) (D9C0); FRNDINT (D9FC);     }
   { FSUBP st(0),st(1) (DEE9); FLDCW mem0 (D96EF4)              }
{$ENDIF}
{ }(Name:'SQR';  Args: 1; Code: #$02#$D8#$C8),
{ }(Name:'CUBE'; Args: 1; Code: #$06#$D9#$C0#$D8#$C8#$DE#$C9),
{$IFDEF WIN32}
{ }(Name:'ROUND';Args: 1; Code: #$08#$DB#$5C#$24#$FC#$DB#$44#$24#$FC),
   { FISTP mem0 (DB5C24FC); FILD mem0 (DB4424FC); }
{$ELSE}
{ }(Name:'ROUND';Args: 1; Code: #$06#$DB#$5E#$F4#$DB#$46#$F4),
   { FISTP mem0 (DB5EF4); FILD mem0 (DB46F4); }
{$ENDIF}
{$IFDEF WIN32}
{ }(Name:'IIF'; Args: 3; Code: #$10#$50#$D9#$E4#$DF#$E0#$9B#$9E#$74#$04 +
                                   #$D9#$C9#$D9#$CA#$DE#$D9#$58),
   { PUSH ax (50); FTST (D9E4); FNSTSW ax (DFE0); FWAIT (9B); SAHF (9E); }
   { JE @CONT (7704); FXCH ST(1) (D9C9); FXCH ST(2) (D9CA);              }
   { @CONT: FCOMPP (DED9); POP ax (58)                                   }
{ }(Name:'SIGN'; Args: 1; Code: #$18#$50#$D9#$E4#$DF#$E0#$9B#$9E#$74#$08 +
                                    #$D9#$E8#$77#$06#$D9#$E0#$EB#$02#$D9#$EE +
                                    #$D9#$C9#$D8#$D9#$58),
   { PUSH ax (50); FTST (D9E4); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E); }
   { JE @EQZERO (7408); FLD1 (D9E8); JA @EXIT (7706); FCHS (D9E0);       }
   { JMP @EXIT (EB02); @EQZERO: FLDZ (D9EE); @MYEXIT: FXCH ST(1) (D9C9); }
   { FCOMP ST(1) (D8D9); POP ax (58) }
{$ELSE}
{ }(Name:'IIF'; Args: 3; Code: #$0E#$D9#$E4#$DF#$E0#$9B#$9E#$74#$04 +
                                   #$D9#$C9#$D9#$CA#$DE#$D9),
   { FTST (D9E4); FNSTSW ax (DFE0); FWAIT (9B); SAHF (9E);    }
   { JE @CONT (7404); FXCH ST(1) (D9C9); FXCH ST(2) (D9CA);   }
   { @CONT: FCOMPP (DED9);                                    }
{ }(Name:'SIGN'; Args: 1; Code: #$16#$D9#$E4#$DF#$E0#$9B#$9E#$74#$08 +
                                    #$D9#$E8#$77#$06#$D9#$E0#$EB#$02#$D9#$EE +
                                    #$D9#$C9#$D8#$D9),
   { FTST (D9E4); FNSTSW AX (DFE0); FWAIT (9B); SAHF (9E);               }
   { JE @EQZERO (7408); FLD1 (D9E8); JA @EXIT (7706); FCHS (D9E0);       }
   { JMP @EXIT (EB02); @EQZERO: FLDZ (D9EE); @MYEXIT: FXCH ST(1) (D9C9); }
   { FCOMP ST(1) (D8D9);                                                 }
{$ENDIF}
{$IFDEF WIN32}
   (Name:'FACT'; Args: 1; Code: #$26#$D9#$E1#$D9#$FC#$DB#$54#$24#$FC#$8B#$4C +
                                    #$24#$FC#$83#$F9#$01#$77#$06#$DD#$D8#$D9#$E8 +
                                    #$EB#$0F#$D9#$E8#$D9#$E8#$D8#$EA#$49#$DC#$CA +
                                    #$D8#$E1#$E2#$FA#$DE#$D9),
   { FABS (D9E1); FRNDINT (D9FC); FIST mem0 (DB5424FC);                     }
   { MOV ecx, mem0 (8B4C24FC); CMP ecx, 1 (83F901); JG @BIG (7F06);         }
   { FSTP ST(0) (DDD8); FLD1 (D9E8); JMP @EXIT (EB0F)                       }
   { @BIG: FLD1 (D9E8); FLD1 (D9E8); FSUBR ST, ST(2) (D8EA); DEC ecx (49)   }
   { @LOOP: FMUL ST(2), ST (DCCA); FSUB ST, ST(1) (D8E1); LOOP @LOOP (E2FA);}
   { FCOMPP (DED9); @EXIT:                                                  }
   (Name:'BINOM'; Args: 2; Code: #$64#$50#$D9#$E1#$D9#$FC#$D9#$C9#$D9#$E1#$D9#$FC +
                                     #$D8#$D1#$9B#$DF#$E0#$9E#$76#$06#$DE#$D9 +
                                     #$D9#$EE#$EB#$4A#$DB#$54#$24#$FC#$8B#$4C#$24#$FC +
                                     #$83#$F9#$01#$77#$10#$0B#$C9#$74#$06#$DD#$D8 +
                                     #$D9#$E8#$EB#$15#$D9#$D9#$D9#$E8#$EB#$2D +
                                     #$D9#$E8#$D9#$E8#$D8#$EA#$49#$DC#$CA#$D8#$E1 +
                                     #$E2#$FA#$DE#$D9#$D9#$C9#$8B#$4C#$24#$FC +
                                     #$83#$F9#$01#$77#$02#$EB#$0F#$D9#$E8#$D9#$E8 +
                                     #$D8#$EA#$49#$DC#$CA#$D8#$E1#$E2#$FA#$DE#$D9#$DE#$F1#$58),
   { PUSH ax (50); FABS (D9E1); FRNDINT (D9FC); FXCH(D9C9); FABS (D9E1);    }
   { FRNDINT (D9FC); FCOM (D8D1); FWAIT (9B); FNSTSW ax (DFE0); SAHF (9E);  }
   { JNA @CONT (7606); FCOMPP (DED9); FLDZ (D9EE); JMP @EXIT (EB4A)         }
   { @CONT: FIST mem0 (DB5424FC); MOV ecx, mem0 (8B4C24FC);                 }
   { CMP ecx, 1 (83F901); JA @BIG1 (7710); OR ecx, ecx (0BC9);              }
   { JZ @EXIT0 (7406); FSTP ST(0) (DDD8); FLD1 (D9E8); JMP @EXIT1 (EB15)    }
   { EXIT0: FCOMPP (D9D9); FLD1 (D9E8); JMP EXIT (EB2D)                     }
   { @BIG1: FLD1 (D9E8); FLD1 (D9E8); FSUBR ST, ST(2) (D8EA); DEC ecx (49)  }
   { @LOOP1: FMUL ST(2), ST (DCCA); FSUB ST, ST(1) (D8E1);                  }
   { LOOP @LOOP1 (E2FA); FCOMPP (DED9); @EXIT1: FXCH (D9C9);                }
   { MOV ecx, mem0 (8B4C24FC); CMP ecx, 1 (83F901); JA @BIG2 (7702)         }
   { JMP @EXIT2 (EB0F)                                                      }
   { @BIG2: FLD1 (D9E8); FLD1 (D9E8); FSUBR ST, ST(2) (D8EA); DEC ecx (49)  }
   { @LOOP2: FMUL ST(2), ST (DCCA); FSUB ST, ST(1) (D8E1);                  }
   { LOOP @LOOP2 (E2FA); FCOMPP (DED9); @EXIT2: FDIVPR (DEF1)               }
   { @EXIT: POP ax (58)                                                     }
{$ELSE}
   (Name:'FACT'; Args: 1; Code: #$24#$D9#$E1#$D9#$FC#$DF#$56#$F4#$8B#$4E#$F4 +
                                    #$83#$F9#$01#$77#$06#$DD#$D8#$D9#$E8 +
                                    #$EB#$0F#$D9#$E8#$D9#$E8#$D8#$EA#$49 +
                                    #$DC#$CA#$D8#$E1#$E2#$FA#$DE#$D9),
   { FABS (D9E1); FRNDINT (D9FC); FIST mem0 (DF56F4);                       }
   { MOV cx, mem0 (8B4EF4); CMP cx, 1 (83F901); JA @BIG (7706)              }
   { FSTP ST(0) (DDD8); FLD1 (D9E8); JMP @EXIT (EB0F)                       }
   { @BIG: FLD1 (D9E8); FLD1 (D9E8); FSUBR ST, ST(2) (D8EA); DEC cx (49)    }
   { @LOOP: FMUL ST(2), ST (DCCA); FSUB ST, ST(1) (D8E1); LOOP @LOOP (E2FA);}
   { FCOMPP (DED9); @EXIT:                                                  }
   (Name:'BINOM'; Args: 2; Code: #$5F#$D9#$E1#$D9#$FC#$D9#$C9#$D9#$E1#$D9#$FC +
                                     #$D8#$D1#$9B#$DF#$E0#$9E#$76#$06#$DE#$D9 +
                                     #$D9#$EE#$EB#$47#$DF#$56#$F4#$8B#$4E#$F4 +
                                     #$83#$F9#$01#$77#$10#$0B#$C9#$74#$06#$DD#$D8 +
                                     #$D9#$E8#$EB#$15#$D9#$D9#$D9#$E8#$EB#$2C +
                                     #$D9#$E8#$D9#$E8#$D8#$EA#$49#$DC#$CA#$D8#$E1 +
                                     #$E2#$FA#$DE#$D9#$D9#$C9#$8B#$4E#$F4 +
                                     #$83#$F9#$01#$77#$02#$EB#$0F#$D9#$E8#$D9#$E8 +
                                     #$D8#$EA#$49#$DC#$CA#$D8#$E1#$E2#$FA#$DE#$D9#$DE#$F1),
   { PUSH ax (50); FABS (D9E1); FRNDINT (D9FC); FXCH(D9C9); FABS (D9E1);    }
   { FRNDINT (D9FC); FCOM (D8D1); FWAIT (9B); FNSTSW ax (DFE0); SAHF (9E);  }
   { JNA @CONT (7606); @ERR: FCOMPP (DED9); FLDZ (D9EE); JMP @EXIT (EB47)   }
   { @CONT: FIST mem0 (DF56F4); MOV cx, mem0 (8B4EF4);                      }
   { CMP cx, 1 (83F901); JA @BIG1 (7710); OR cx, cx (0BC9);                 }
   { JZ @EXIT0 (7406); FSTP ST(0) (DDD8); FLD1 (D9E8); JMP @EXIT1 (EB15)    }
   { EXIT0: FCOMPP (D9D9); FLD1 (D9E8); JMP EXIT (EB2C)                     }
   { @BIG1: FLD1 (D9E8); FLD1 (D9E8); FSUBR ST, ST(2) (D8EA); DEC cx (49)   }
   { @LOOP1: FMUL ST(2), ST (DCCA); FSUB ST, ST(1) (D8E1);                  }
   { LOOP @LOOP1 (E2FA); FCOMPP (DED9); @EXIT1: FXCH (D9C9);                }
   { MOV cx, mem0 (8B4EF4); CMP cx, 1 (83F901); JA @BIG2 (7702)             }
   { JMP @EXIT2 (EB0F)                                                      }
   { @BIG2: FLD1 (D9E8); FLD1 (D9E8); FSUBR ST, ST(2) (D8EA); DEC cx (49)   }
   { @LOOP2: FMUL ST(2), ST (DCCA); FSUB ST, ST(1) (D8E1);                  }
   { LOOP @LOOP2 (E2FA); FCOMPP (DED9); @EXIT2: FDIVPR (DEF1) @EXIT:        }
{$ENDIF}
   (Name:'COSH'; Args: 1; Code: #$2A#$D9#$EA#$DE#$C9#$D9#$C0#$D9#$FC +
                                #$DC#$E9#$D9#$E8#$D9#$FD#$DD#$D9 +
                                #$D9#$C9#$D9#$F0#$D9#$E8#$DE#$C1 +
                                #$DE#$C9#$D9#$C0#$D9#$E8#$DE#$F1 +
                                #$DE#$C1#$D9#$E8#$D9#$E8#$DE#$C1#$DE#$F9),
   { FLDL2E (D9EA); FMULP ST(1),ST(0) (DEC9); FLD ST(0) (D9C0);        }
   { FRNDINT (D9FC); FSUB ST(1),ST(0) (DCE9) FLD1 (D9E8);              }
   { FSCALE (D9FD); FSTP ST(1) (DDD9); FXCH ST(1) (D9C9);              }
   { F2XM1 (D9F0); FLD1 (D9E8); FADDP ST(1),ST(0) (DEC1);              }
   { FMULP ST(1),ST(0) (DEC9); FLD ST(0) (D9C0); FLD1 (D9E8);          }
   { FDIVR (DEF1); FADD (DEC1); FLD1(D9E8); FLD1(D9E8); FADD (DEC1);   }
   { FDIV (DEF9) }
   (Name:'SINH'; Args: 1; Code: #$2A#$D9#$EA#$DE#$C9#$D9#$C0#$D9#$FC +
                                #$DC#$E9#$D9#$E8#$D9#$FD#$DD#$D9 +
                                #$D9#$C9#$D9#$F0#$D9#$E8#$DE#$C1 +
                                #$DE#$C9#$D9#$C0#$D9#$E8#$DE#$F1 +
                                #$DE#$E9#$D9#$E8#$D9#$E8#$DE#$C1#$DE#$F9),
   { FLDL2E (D9EA); FMULP ST(1),ST(0) (DEC9); FLD ST(0) (D9C0);        }
   { FRNDINT (D9FC); FSUB ST(1),ST(0) (DCE9) FLD1 (D9E8);              }
   { FSCALE (D9FD); FSTP ST(1) (DDD9); FXCH ST(1) (D9C9);              }
   { F2XM1 (D9F0); FLD1 (D9E8); FADDP ST(1),ST(0) (DEC1);              }
   { FMULP ST(1),ST(0) (DEC9); FLD ST(0) (D9C0); FLD1 (D9E8);          }
   { FDIVR (DEF1); FSUB (D9E9); FLD1(D9E8); FLD1(D9E8); FADD (DEC1);   }
   { FDIV (DEF9) }
   (Name:'TANH'; Args: 1; Code: #$26#$D9#$EA#$DE#$C9#$D9#$C0#$D9#$FC +
                                #$DC#$E9#$D9#$E8#$D9#$FD#$DD#$D9 +
                                #$D9#$C9#$D9#$F0#$D9#$E8#$DE#$C1 +
                                #$DE#$C9#$D8#$C8#$D9#$C0#$D9#$E8 +
                                #$DC#$C2#$DE#$E9#$DE#$F1),
   { FLDL2E (D9EA); FMULP ST(1),ST(0) (DEC9); FLD ST(0) (D9C0);        }
   { FRNDINT (D9FC); FSUB ST(1),ST(0) (DCE9) FLD1 (D9E8);              }
   { FSCALE (D9FD); FSTP ST(1) (DDD9); FXCH ST(1) (D9C9);              }
   { F2XM1 (D9F0); FLD1 (D9E8); FADDP ST(1),ST(0) (DEC1);              }
   { FMULP ST(1),ST(0) (DEC9); FMUL ST, ST(0) (D8C8); FLD ST(0) (D9C0) }
   { FLD1 (D9E8); FADD ST(2), ST (DCC2);  FSUBP ST(1), ST (DEE9);      }
   { FDIVRP (DEF1)                                                     }
   (Name:'ACOSH'; Args: 1; Code: #$12#$D9#$C0#$D8#$C8#$D9#$E8#$DE +
                                     #$E9#$D9#$FA#$DE#$C1#$D9#$ED +
                                     #$D9#$C9#$D9#$F1),
   { FLD ST(0) (D9C0); FMUL ST,ST(0) (D8C8); FLD1 (D9E8);              }
   { FSUBP  ST(1),ST (DEE9); FSQRT (D9FA); FADDP  ST(1),ST (DEC1);     }
   { FLDLN2 (D9ED); FXCH ST(1) (D9C9); FYL2X (D9F1)                    }
   (Name:'ASINH'; Args: 1; Code: #$12#$D9#$C0#$D8#$C8#$D9#$E8#$DE +
                                     #$C1#$D9#$FA#$DE#$C1#$D9#$ED +
                                     #$D9#$C9#$D9#$F1),
   { FLD ST(0) (D9C0); FMUL ST,ST(0) (D8C8); FLD1 (D9E8);              }
   { FADDP  ST(1),ST (DEC1); FSQRT (D9FA); FADDP  ST(1),ST (DEC1);     }
   { FLDLN2 (D9ED); FXCH ST(1) (D9C9); FYL2X (D9F1)                    }
{ }(Name:'ATANH'; Args: 1; Code: #$14#$D9#$C0#$D9#$E8#$DC#$E2#$DE +
                                     #$C1#$DE#$F1#$D9#$FA#$D9#$ED#$D9#$C9 +
                                     #$D9#$F1),
   { FLD ST(0) (D9C0); FLD1 (D9E8); FSUBR  ST(2),ST (DCE2);            }
   { FADDP  ST(1),ST (DEC1); FDIVRP  ST(1),ST (DEF1); FSQRT (D9FA)     }
   { FLDLN2 (D9ED);  FXCH   ST(1) (D9C9); FYL2X (D9F1);                }
{$IFDEF WIN32}
{ }(Name:'MIN'; Args: -1; Code: #$0E#$50#$D8#$D1#$9B#$DF#$E0#$9E#$77#$02 +
                                    #$D9#$C9#$DD#$D8#$58),
   { PUSH ax (50); FCOM (D8D1); FWAIT (9B); FNSTSW ax (DFE0); SAHF (9E);  }
   { JA @CONT (7702); FXCH ST(1) (D9C9); @CONT: FSTP (DDD8); POP ax (58); }
{ }(Name:'MAX'; Args: -1; Code: #$0E#$50#$D8#$D1#$9B#$DF#$E0#$9E#$72#$02 +
                                    #$D9#$C9#$DD#$D8#$58),
   { PUSH ax (50); FCOM (D8D1); FWAIT (9B); FNSTSW ax (DFE0); SAHF (9E);  }
   { JB @CONT (7202); FXCH ST(1) (D9C9); @CONT: FSTP (DDD8); POP ax (58); }
{$ELSE}
{ }(Name:'MIN'; Args: -1; Code: #$0C#$D8#$D1#$9B#$DF#$E0#$9E#$77#$02 +
                                    #$D9#$C9#$DD#$D8),
   { FCOM (D8D1); FWAIT (9B); FNSTSW ax (DFE0); SAHF (9E);   }
   { JA @CONT (7702); FXCH ST(1) (D9C9); @CONT: FSTP (DDD8); }
{ }(Name:'MAX'; Args: -1; Code: #$0C#$D8#$D1#$9B#$DF#$E0#$9E#$72#$02 +
                                    #$D9#$C9#$DD#$D8),
   { FCOM (D8D1); FWAIT (9B); FNSTSW ax (DFE0); SAHF (9E);   }
   { JB @CONT (7202); FXCH ST(1) (D9C9); @CONT: FSTP (DDD8); }
{$ENDIF}
{ }(Name:'SUM'; Args: -1; Code: #$02#$DE#$C1),
   { FADDP (DEC1)                                               }
{ }(Name:'PROD'; Args: -1; Code: #$02#$DE#$C9)
   { FADDP (DEC9)                                               }
);

{===============================================================}
{ Comments:                                                     }
{ for 32-bit           for 16-bit                               }
{ mem0 = [esp-4]       mem0 = ss:[sp-12]                        }
{ mem2 = [esp-8]       mem2 = ss:[sp-10]                        }
{                                                               }
{ * - means that item can't be moved or deleted                 }
{===============================================================}

implementation

{ Error handler }
procedure FCError(ErrorCode: Integer); {$IFNDEF WIN32} far; {$ENDIF}
begin
  if ErrorCode <> fcOk then
    if ErrorCode > 0 then
      raise EFCError.CreateFmt('Syntax error at pos %d', [ErrorCode])
    else case ErrorCode of
      fcInvalidId: raise EFCError.Create('Invalid identifier');
      fcReservedId: raise EFCError.Create('Reserved identifier');
      fcUsedName: raise EFCError.Create('This entry is referenced by others');
      fcUnknownVar: raise EFCError.Create('Unknown variable name');
      fcUnknownFunc: raise EFCError.Create('Unknown function name');
      else raise EFCError.Create('Internal error');
    end;
end;

{ Check if a Source contains identifier Ident }
function ContainsIdent(const Source, Ident: String): Integer;
var
  tS, tI: String;
  tB: Boolean;
begin
  tS := UpperCase(Source);
  tI := UpperCase(Ident);
  repeat
    Result := Pos(tI, tS);
    tB := Result > 0;
    if tB
      then begin
        if Result > 1 then tB := tB and not (tS[Result-1] in Numbers + Letters);
        if Result + Length(tI) <= Length(tS)
          then tB := tB and not (tS[Result+Length(tI)] in Numbers + Letters);
        tS[Result+Length(tI)-1] := '#';
      end
      else break;
  until tB;
  if not tB then Result := -1;
end;

{ Get number of arguments in S , Lib - user aliases library }
function Arguments(const S: String; Lib: TFormulaLib): Integer;
var
  Temp: String;
  i, j, tI, ErrorFlag: Integer;
  tS: String[7];
  tB: Boolean;
begin
  Result := -1;
  if Lib <> nil
    then Temp := Lib.ExpandAliases(S)
    else Temp := S;
  i := 1;
  while i <= Length(Temp) do
  begin
    { Check for %n }
    if Temp[i] = ArgumentMarker
    then begin
      tS := '';
      Inc(i);
      while ( i <= Length(Temp) ) and ( Temp[i] in Numbers ) do
      begin
        tS := tS + Temp[i];
        Inc(i);
      end;
      Val(tS, tI, ErrorFlag);
      if ( ErrorFlag = 0 ) and ( Result < tI ) then Result := tI;
    end;
    Inc(i);
  end;
  Result := Result + 1; { Because the first argument if 0th }
end;

{ Checks the validity of function or constant name given in S }
{ Merely look up for S in the BuiltInFuncs }
function ValidBIFName(const S: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to BIFCount - 1 do
    if CompareText(S, BuiltInFuncs[i].Name) = 0
    then begin
      Result := i;
      break;
    end;
end;

{ Checks the validity of logic operator given in S }
{ Merely look up for S in the BuiltInOperators }
function ValidLogicOpName(const S: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to LogicOpCount - 1 do
    if CompareText(S, BuiltInOperators[i].Name) = 0
    then begin
      Result := i;
      break;
    end;
end;

{ Parse S and put its internal representation into InternalExpr }
function MakeInternalExpr(S: String; InternalExpr: TInternalExprArray; NameList: TStringList; Lib: TFormulaLib): Integer;
var
  LastElem: Char;
  tS: String[31];

{ Processes argument list of multyargument function }
{ i - position in S, Args: number of arguments (should be), RealArg: real
{ number of arguments }
{ Returns TRUE if processing was successful, FALSE overwise }
function ProcessArgsList(var i: Integer; Args: Integer; var RealArgs: Integer): Boolean;

{ ProcessLetters - procees part of S which begins with letter }
{ Example: S = '4+cos(2)' -  ProcessLetter will be called (i = 3, S[i] = 'c') }
function ProcessLetters(var i: Integer): Boolean;
var
  FN, RA, AmpPlace: Integer;
begin
  ProcessLetters := True;
  { store function or constant name in tS }
  tS := '';
  while ( S[i] in Letters + Numbers ) and ( i <= Length(S) ) do
  begin
    tS := tS + S[i];
    Inc(i);
  end;
  Dec(i);
  FN := ValidBIFName(tS);
  { if tS is BuiltInFunc }
  if FN >= 0
  then begin
    if BuiltInFuncs[FN].Args <> 0
    then begin
      InternalExpr.Add(MakeItem('f', FN, 0));
      AmpPlace := InternalExpr.Count;
      InternalExpr.Add(MakeItem('&', BuiltInFuncs[FN].Args, 0));
      ProcessLetters := ProcessArgsList(i, BuiltInFuncs[FN].Args, RA);
      if RA <> BuiltInFuncs[FN].Args
        then begin
          InternalExpr.Cnst[AmpPlace] := 1;
          InternalExpr.Id[AmpPlace] := RA;
        end;
    end
    else InternalExpr.Add(MakeItem('c', FN, 0));
    exit;
  end;
  if Lib <> nil
    then begin
      FN := Lib.ValidFunctionName(tS);
      { if tS is user function or constant }
      if FN >= 0
      then begin
        if Lib.Functions.ArgCount[FN] > 0
        then begin
          InternalExpr.Add(MakeItem('f', FN + UserFuncBase, 0));
          InternalExpr.Add(MakeItem('&', Lib.Functions.ArgCount[FN], 0));
          ProcessLetters := ProcessArgsList(i, Lib.Functions.ArgCount[FN], RA);
        end
        else InternalExpr.Add(MakeItem('c', FN + UserFuncBase, 0));
        exit;
      end;
    end;
  if NameList <> nil
  then begin
    if S[i+1] = '('
      then begin
        InternalExpr.Add(MakeItem('u', NameList.Add(tS), 0));
        AmpPlace := InternalExpr.Count;
        InternalExpr.Add(MakeItem('&', 0, 0));
        ProcessLetters := ProcessArgsList(i, -1, RA);
        InternalExpr.Id[AmpPlace] := RA;
      end
      else
        InternalExpr.Add(MakeItem('v', NameList.Add(tS), 0));
    exit;
  end;
  ProcessLetters := False;
end;

{ ProcessNumbers - process part of S which begins with number }
{ Example: S = '4+cos(2)' -  ProcessLetter will be called (i = 1, S[i] = '4') }
function ProcessNumbers(var i: Integer): Boolean;
var
  N: Extended;
  erCode: Integer;
begin
  ProcessNumbers := False;
  tS := '';
  { store number in tS }
  while ( ( S[i] in Numbers + ['e', 'E', '+', '-'])
         or ( S[i] = DecimalSeparator ) ) and
        ( i <= Length(S) ) do
  begin
    if (S[i] in ['+', '-']) and not (S[i-1] in ['e', 'E'])
      then Break;
    tS := tS + S[i];
    Inc(i);
  end;
  Dec(i);
  { convert tS in numeric representation }
  Val(tS, N, erCode);
  { if convertion succesfull }
  if erCode = 0
  then begin
    ProcessNumbers := True;
    InternalExpr.Add(MakeItem('c', -1, N));
  end;
end;

{ ProcessLogicOperator - procees part of S which begins with <, >, =, ... }
{ Example: S = '4<>cos(2)' -  ProcessLogicOperator will be called  }
{ (i = 2, S[i] = '<') }
function ProcessLogicOperator(var i: Integer): Boolean;
var
  N, erCode: Integer;
begin
  ProcessLogicOperator := False;
  tS := '';
  { store argument number in tS }
  while ( i < Length(S) ) and ( S[i] in Operators ) do
  begin
    tS := tS + S[i];
    Inc(i);
  end;
  Dec(i);
  { check if tS is valid operator }
  N := ValidLogicOpName(tS);
  if N >= 0
  then begin
    ProcessLogicOperator := True;
    InternalExpr.Add(MakeItem('o', N, 0));
  end;
end;

{ ProcessArgument - procees part of S which begins with '%' }
{ Example: S = '%0+2' -  ProcessArgument will be called (i = 1, S[i] = '%') }
function ProcessArgument(var i: Integer): Boolean;
var
  N, erCode: Integer;
begin
  ProcessArgument := False;
  tS := '';
  { store argument number in tS }
  while ( i < Length(S) ) and ( S[i+1] in Numbers ) do
  begin
    Inc(i);
    tS := tS + S[i];
  end;
  { convert tS in number }
  Val(tS, N, erCode);
  { if conversion succesfull }
  if erCode = 0
  then begin
    ProcessArgument := True;
    InternalExpr.Add(MakeItem('a', N, 0));
  end;
end;

{ Checks the validity of the combination in the list}
function Combination(E1, E2: Char): Boolean;
begin
  Combination :=
    ( (E1 in ['(', ArgumentSeparator] ) and ( E2 in ['(', 'a', 'c', 'f', '-', '+', '!', 'u', 'v'] ) ) or
    ( (E1 in Operations + ['o'] ) and ( E2 in ['(', 'a', 'c', 'f', 'u', 'v'] ) ) or
    ( (E1 in ['a', 'c', 'v', ')']) and ( E2 in [')', ArgumentSeparator, 'o'] + Operations - ['!']) or
    ( (E1 in ['f', 'u']) and ( E2 = '&' ) ) or
    ( (E1 = '&') and ( E2 = '(' ) ));
end;

var
  Braces: Integer;
  Valid: Boolean;
begin
  Valid := True;
  { Internal variable, it stores differences between '(' and ')' }
  Braces := 0;
  { Number of arguments in the function, later it will be compared with Args }
  RealArgs := 1;
  { Main cycle we will process list of arguments in the brackets }
  repeat
    Inc(i);
    { case depends on S[i] }
    if S[i] = ArgumentMarker
      then Valid := ProcessArgument(i)
      else
        if S[i] in Numbers
        then Valid :=  ProcessNumbers(i)
        else
          if S[i] in Letters
          then Valid := ProcessLetters(i)
          else
            if S[i] in Operators
            then Valid := ProcessLogicOperator(i)
            else
              if S[i] = ArgumentSeparator
              then begin
                Inc(RealArgs);
                InternalExpr.Add(MakeItem(S[i], 0, 0))
              end
              else
                if S[i] in [' ', #10, #13]
                  then Continue
                  else begin
                    InternalExpr.Add(MakeItem(S[i], 0, 0));
                    case S[i] of
                     '(': Inc(Braces);
                     ')': Dec(Braces);
                    end;
                  end;
    { Valid - logical value means whether formula is valid or not }
    Valid := Valid and (Braces >= 0) and
             ((Args < 0) or (RealArgs <= Args )) and
             Combination(LastElem, InternalExpr.Elem[InternalExpr.Count-1]);
    LastElem := InternalExpr.Elem[InternalExpr.Count - 1];
  until ( Braces = 0 ) or ( i = Length(S) ) or ( not Valid );
  ProcessArgsList := Valid and ((Args < 0) or (Args = RealArgs)) and
                     ( LastElem = ')' ) and ( Braces = 0 );
end;

var
  i, tA: Integer;
begin
  if InternalExpr <> nil
    then begin
      if S = ''
        then S := '0'
        else if Lib <> nil
               then S := '(' + Lib.ExpandAliases(S) + ')'
               else S := '(' + S + ')';
      LastElem := '(';
      i := 0;
      if ProcessArgsList(i, 1, tA) and ( i = Length(S) )
        then Result := fcOk else Result := i - 1;
     end
     else Result := fcInternal;
end;

procedure Optimize(InternalExpr: TInternalExprArray);
var
  i, j, Brackets: Integer;
begin
  if InternalExpr <> nil
  then
    with InternalExpr do
    begin
      for i := Count - 1 downto 0 do
      begin
        { Convert Unary "-","+","!" ( it can be a) first; b) after "(" and "," ) }
        if ( Elem[i] = '+' ) and
           ( ( i = 0 ) or ( Elem[i-1] = '(' ) or ( Elem[i-1] = ArgumentSeparator ) )
        then Delete(i);
        if ( Elem[i] = '!' ) and
           ( ( i = 0 ) or ( Elem[i-1] = '(' ) or ( Elem[i-1] = ArgumentSeparator ) )
        then begin
            Delete(i);
            Insert(i, MakeItem('f', 5, 0));
            Insert(i+1, MakeItem('&', 1, 0));
            case Elem[i+2] of
            'f', 'u': begin
              Insert(i+2, MakeItem('(', 0, 0));
              Brackets := 1;
              j := i + 5;
              while Brackets > 0 do
              begin
                if Elem[j] = ')' then Dec(Brackets);
                if Elem[j] = '(' then Inc(Brackets);
                Inc(j);
              end;
              Insert(j, MakeItem(')', 0, 0));
            end;
            'a','c','v': begin
              Insert(i+2, MakeItem('(', 0, 0));
              Insert(i+4, MakeItem(')', 0, 0));
            end;
            end;
          end;
        if ( Elem[i] = '-' ) and
           ( ( i = 0 ) or ( Elem[i-1] = '(' ) or ( Elem[i-1] = ArgumentSeparator ) )
        then
          { Before constant }
          if ( Elem[i+1] = 'c' ) and ( Id[i+1] = -1 )
          then begin
            Cnst[i+1] := -Cnst[i+1];
            Delete(i);
          end
          { before arument or function }
          else begin
            Delete(i);
            Insert(i, MakeItem('f', 0, 0));
            Insert(i+1, MakeItem('&', 1, 0));
            case Elem[i+2] of
            'f', 'u': begin
              Insert(i+2, MakeItem('(', 0, 0));
              Brackets := 1;
              j := i + 5;
              while Brackets > 0 do
              begin
                if Elem[j] = ')' then Dec(Brackets);
                if Elem[j] = '(' then Inc(Brackets);
                Inc(j);
              end;
              Insert(j, MakeItem(')', 0, 0));
            end;
            'a','c','v': begin
              Insert(i+2, MakeItem('(', 0, 0));
              Insert(i+4, MakeItem(')', 0, 0));
            end;
            end;
          end;
        { 0 to FLDZ and 1 to FLD1 }
        if (Elem[i] = 'c') and (Id[i] = -1) and (Cnst[i] = 0.0)
          then Id[i] := 1;
        if (Elem[i] = 'c') and (Id[i] = -1) and (Cnst[i] = 1.0)
          then Id[i] := 2;
      end; {for}
    end; {with}
end;

{ Makes RPN from InternalExpr and put it to Postfix }
{ It is a modification of well known algorithm to convert usual expression }
{ to Reverse Polish Notaton (RPN) }
procedure MakePostfix(InternalExpr, Postfix: TInternalExprArray);
var
  SStack: TInternalExprArray;
  i: Integer;

{ Defines the Priority }
function Prior(Ch: char): Byte;
begin
  case Ch of
    '(': Prior := 1;
    ')': Prior := 2;
    ArgumentSeparator: Prior := 3;
    'o': Prior := 4;
    '+', '-': Prior := 5;
    '*', '/': Prior := 6;
    '^': Prior := 7;
    '&': Prior := 8;
  else Prior := 0;
  end;
end;

begin
  if Postfix <> nil
  then begin
    SStack := TInternalExprArray.Create;
    for i := 0 to InternalExpr.Count - 1 do
      case Prior(InternalExpr.Elem[i]) of
        0: Postfix.Add(InternalExpr.Items[i]);
        1: SStack.Add(InternalExpr.Items[i]);
        else begin
          while ( SStack.Count > 0 ) and
                ( Prior(SStack.Elem[SStack.Count - 1]) >= Prior(InternalExpr.Elem[i]) ) do
          begin
            Postfix.Add(SStack.Items[SStack.Count - 1]);
            SStack.Delete(SStack.Count - 1);
          end;
          if InternalExpr.Elem[i] in Operations + ['&', 'o']
            then SStack.Add(InternalExpr.Items[i])
          else if InternalExpr.Elem[i] = ')'
               then SStack.Delete(SStack.Count - 1);
        end; { else }
      end; { case }
    SStack.Free;
  end; { if }
end;

{ TArray methods }
constructor TArray.Create;
begin
  inherited Create;
  FItemSize := AItemSize;
end;

destructor TArray.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TArray.Assign(AArray: TArray);
begin
  if ClassType = AArray.ClassType
  then begin
    Clear;
    SetCapacity(AArray.FCapacity);
    Move(AArray.FArray^, FArray^, AArray.FCount*FItemSize);
    FCount := AArray.FCount;
  end
  else Error(aTypecast);
end;

function TArray.Add(const Item): Integer;
begin
  Result := FCount;
  if Result = FCapacity then Grow;
  Move(Item, FArray[FCount*FItemSize], FItemSize);
  Inc(FCount);
end;

procedure TArray.GetItem(Index: Integer; var Item);
begin
  if (Index < 0) or (Index >= FCount) then Error(aIndex);
  Move(FArray[Index*FItemSize], Item, FItemSize);
end;

procedure TArray.SetItem(Index: Integer; const Item);
begin
  if (Index < 0) or (Index >= FCount) then Error(aIndex);
  Move(Item, FArray[Index*FItemSize], FItemSize);
end;

procedure TArray.Clear;
begin
  FCount := 0;
  SetCapacity(0);
end;

procedure TArray.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then Error(aIndex);
  Dec(FCount);
  if Index < FCount then
    Move(FArray[(Index + 1)*FItemSize], FArray[Index*FItemSize],
      (FCount - Index) * FItemSize);
end;

procedure TArray.Error(ErrorCode: Integer);
begin
  raise Exception.Create('TArray error. Code ' + IntToStr(ErrorCode));
end;

procedure TArray.Grow;
begin
  SetCapacity(FCapacity + 8);
end;

procedure TArray.Insert(Index: Integer; const Item);
begin
  if (Index < 0) or (Index > FCount) then Error(aIndex);
  if FCount = FCapacity then Grow;
  if Index < FCount then
    Move(FArray[Index*FItemSize], FArray[(Index + 1)*FItemSize],
      (FCount - Index) * FItemSize);
  Move(Item, FArray[Index*FItemSize], FItemSize);
  Inc(FCount);
end;

function TArray.IndexOf(const Item): Integer;

function MemComp(Mem1, Mem2: PChar; Length: Cardinal): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Length - 1 do
    if Mem1[i] <> Mem2[i] then exit;
  Result := True;
end;

var
  i: Integer;
begin
  Result := -1;
  for i := 0 to FCount - 1 do
    if MemComp(FArray + i*FItemSize, @Item, FItemSize)
      then begin
        Result := i;
        exit;
      end;
end;

procedure TArray.SetCapacity(NewCapacity: Integer);
begin
  if (NewCapacity < FCount) or (NewCapacity*FItemSize > MaxArraySize) then Error(aCapacity);
  if NewCapacity <> FCapacity then
  begin
    {$IFDEF WIN32}
    ReallocMem(FArray, NewCapacity * FItemSize);
    {$ELSE}
    FArray := ReallocMem(FArray, FCapacity * FItemSize, NewCapacity * FItemSize);
    {$ENDIF}
    FCapacity := NewCapacity;
  end;
end;

function MakeItem(AElem: Char; AID: Integer; ACnst: Extended): TInternalExprItem;
begin
  with Result do
  begin
    Elem := AElem;
    Id := AId;
    Cnst := ACnst;
  end;
end;

{ TInternalExprArray methods }
constructor TInternalExprArray.Create;
begin
  inherited Create(SizeOf(TInternalExprItem));
end;

procedure TInternalExprArray.Add(Item: TInternalExprItem);
begin
  inherited Add(Item);
end;

procedure TInternalExprArray.Insert(Index: Integer; Item: TInternalExprItem);
begin
  inherited Insert(Index, Item);
end;

function TInternalExprArray.GetElem(Index: Integer): Char;
var
  Temp: TInternalExprItem;
begin
  GetItem(Index, Temp);
  Result := Temp.Elem;
end;

function TInternalExprArray.GetId(Index: Integer): Integer;
var
  Temp: TInternalExprItem;
begin
  GetItem(Index, Temp);
  Result := Temp.Id;
end;

function TInternalExprArray.GetCnst(Index: Integer): Extended;
var
  Temp: TInternalExprItem;
begin
  GetItem(Index, Temp);
  Result := Temp.Cnst;
end;

function TInternalExprArray.Get(Index: Integer): TInternalExprItem;
begin
  GetItem(Index, Result);
end;

procedure TInternalExprArray.SetElem(Index: Integer; NewValue: Char);
var
  Temp: TInternalExprItem;
begin
  GetItem(Index, Temp);
  Temp.Elem := NewValue;
  SetItem(Index, Temp);
end;

procedure TInternalExprArray.SetId(Index: Integer; NewValue: Integer);
var
  Temp: TInternalExprItem;
begin
  GetItem(Index, Temp);
  Temp.Id := NewValue;
  SetItem(Index, Temp);
end;

procedure TInternalExprArray.SetCnst(Index: Integer; NewValue: Extended);
var
  Temp: TInternalExprItem;
begin
  GetItem(Index, Temp);
  Temp.Cnst := NewValue;
  SetItem(Index, Temp);
end;

{ TCodeArray methods }
constructor TCodeArray.Create;
begin
  inherited Create(1);
end;

procedure TCodeArray.Add(AB: Byte);
begin
  inherited Add(AB);
end;

procedure TCodeArray.AddInteger(AL: Integer);
var
  Dummy: array[0..SizeOf(Integer)-1] of byte absolute AL;
  I: Integer;
begin
  for i := 0 to SizeOf(Integer)-1 do
    Add(Dummy[i]);
end;

procedure TCodeArray.AddCode(AC: PChar);
var
  I: Integer;
begin
  for I := 1 to Byte(AC[0]) do
    Add(byte(AC[I]));
end;

procedure TCodeArray.InsertInteger(Index: Integer; AL: Integer);
var
  Dummy: array[0..SizeOf(Integer)-1] of byte absolute AL;
  I: Integer;
begin
  for i := SizeOf(Integer)-1 downto 0 do
    Insert(Index, Dummy[i]);
end;

function TCodeArray.GetInteger(Index: Integer): Integer;
var
  Dummy: array[0..SizeOf(Integer)-1] of byte absolute Result;
  I: Integer;
begin
  for i := SizeOf(Integer)-1 downto 0 do
    GetItem(Index + i, Dummy[i]);
end;

procedure TCodeArray.InsertCode(Index: Integer; AC: PChar);
var
  I: Integer;
begin
  for I := 1 to Byte(AC[0]) do
    Insert(Index, AC[I]);
end;

{ TIntArray methods }
constructor TIntArray.Create;
begin
  inherited Create(SizeOf(Integer));
end;

procedure TIntArray.Add;
begin
  inherited Add(AL);
end;

function TIntArray.GetInt(Index: Integer): Integer;
begin
  GetItem(Index, Result);
end;

procedure TIntArray.SetInt(Index: Integer; NewValue: Integer);
begin
  SetItem(Index, NewValue);
end;

initialization
  @FCCheck := @FCError;
end.
