unit BigNode;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
    String10 = String[10];
    TBigNode = class(TObject)
        public
            node_value : String10;
            children   : TList;
            position   : TRect;

            constructor Create;
            destructor Destroy; override;
            procedure SetPosition(var start_x : Integer; start_y : Integer);
            procedure DrawNode(cvs : TCanvas; selected_node : TBigNode);
            procedure DrawSubtree(cvs : TCanvas; selected_node : TBigNode);
            function NodeAtPoint(X, Y : Integer) : TBigNode;
            function RemoveNode(target_node : TBigNode) : Boolean;
            function FindNodeValue(target_value : String10) : TBigNode;
    end;

  TBigTreeForm = class(TForm)
    Label1: TLabel;
    txtValue: TEdit;
    cmdAdd: TButton;
    cmdRemove: TButton;
    cmdFind: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure txtValueChange(Sender: TObject);
    procedure cmdAddClick(Sender: TObject);
    procedure cmdRemoveClick(Sender: TObject);
    procedure cmdFindClick(Sender: TObject);
  private
    { Private declarations }
    selected_node : TBigNode;

    procedure SelectNode(new_selected : TBigNode);

  public
    { Public declarations }
  end;

var
  BigTreeForm: TBigTreeForm;

implementation

{$R *.DFM}

const
    BOX_WID = 40;
    BOX_HGT = 16;
    BOX_HGAP = 2;
    BOX_VGAP = 6;
var
    root                 : TBigNode;
    tree_xmin, tree_ymin : Integer;

// **************************
// * TBigTreeForm functions *
// **************************
// Create the root node.
procedure TBigTreeForm.FormCreate(Sender: TObject);
var
    start_x : Integer;
begin
    // Create the root node.
    root := TBigNode.Create;
    root.node_value := 'Root';

    // Start with no node selected.
    selected_node := nil;

    // Save the coordinates where we will start drawing.
    tree_xmin := 3;
    tree_ymin := cmdAdd.Top + cmdAdd.Height + 3;

    // Use bold text.
    Canvas.Font.Style := [fsBold];

    // Position the root.
    start_x := tree_xmin;
    root.SetPosition(start_x, tree_ymin);
end;

// Draw the tree.
procedure TBigTreeForm.FormPaint(Sender: TObject);
var
    rect : TRect;
begin
    // Erase the canvas.
    rect.Left := 0;
    rect.Right := Width;
    rect.Top := 0;
    rect.Bottom := Height;
    Canvas.Brush.Color := clLtGray;
    Canvas.FillRect(rect);

    // Draw the tree.
    root.DrawSubtree(Canvas, selected_node);
end;

// Select the node the user clicked.
procedure TBigTreeForm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
    SelectNode(root.NodeAtPoint(X, Y));
end;

// Unselect the previously selected
// node and select this node.
procedure TBigTreeForm.SelectNode(new_selected : TBigNode);
begin
    // Unselect the previously selected node.
    if (selected_node <> nil) then
        selected_node.DrawNode(Canvas, nil);

    // Save the newly selected node.
    selected_node := new_selected;

    // Highlight the selected node.
    if (selected_node <> nil) then
        selected_node.DrawNode(Canvas, selected_node);

    // If there is a node selected and the value TextBox
    // is not empty, enable the Add button.
    cmdAdd.Enabled :=
        ((selected_node <> nil) and
         (Trim(txtValue.Text) <> ''));

    // Enable the Remove button if a node other than
    // the root is selected.
    cmdRemove.Enabled :=
        ((selected_node <> nil) and
         (selected_node <> root));
end;

// Enable and disable buttons appropriately.
procedure TBigTreeForm.txtValueChange(Sender: TObject);
begin
    cmdAdd.Enabled :=
        ((selected_node <> nil) and
         (Trim(txtValue.Text) <> ''));
    cmdFind.Enabled := (Trim(txtValue.Text) <> '');
end;

// Add a child to the selected node.
procedure TBigTreeForm.cmdAddClick(Sender: TObject);
var
    child   : TBigNode;
    start_x : Integer;
begin
    // Create the child.
    child := TBigNode.Create;
    child.node_value := txtValue.Text;

    // Add the child to the selected node's children list.
    selected_node.children.Add(child);

    // Reposition the nodes.
    start_x := tree_xmin;
    root.SetPosition(start_x, tree_ymin);

    // Redraw the tree.
    Refresh;

    // Erase the value TextBox and give it focus.
    txtValue.Text := '';
    txtValue.SetFocus;
end;

// Remove the selected node.
procedure TBigTreeForm.cmdRemoveClick(Sender: TObject);
var
    start_x : Integer;
begin
    // Remove the node.
    root.RemoveNode(selected_node);

    // Deselect the node.
    selected_node := nil;
    cmdRemove.Enabled := False;

    // Reposition the remaining nodes.
    start_x := tree_xmin;
    root.SetPosition(start_x, tree_ymin);

    // Redraw the tree.
    Refresh;
end;

// Find the node with the entered text.
procedure TBigTreeForm.cmdFindClick(Sender: TObject);
var
    target_node : TBigNode;
begin
    // Find the node.
    target_node := root.FindNodeValue(txtValue.Text);

    // Select it.
    SelectNode(target_node);

    // If the node is not here, beep.
    if (target_node = nil) then Beep;
end;

// **********************
// * TBigNode functions *
// **********************
// Create the children list.
constructor TBigNode.Create;
begin
    inherited Create;
    children := TList.Create;
end;

// Free any children and the children list.
destructor TBigNode.Destroy;
var
    i     : Integer;
    child : TBigNode;
begin
    // Free the children.
    for i := 0 to children.Count - 1 do
    begin
        child := children.Items[i];
        child.Free;
    end;

    // Free the children list.
    children.Free;
    inherited Destroy;
end;

// Position the node and its descendants. Update start_x
// so it indicates the rightmost position used by the
// node and its descendants.
procedure TBigNode.SetPosition(var start_x : Integer; start_y : Integer);
var
    i, xmin : Integer;
    child   : TBigNode;
begin
    // Set the node's top and bottom.
    position.Top := start_y;
    position.Bottom := start_y + BOX_HGT;

    // Record the leftmost position used.
    xmin := start_x;

    // If there are no children, put the node here.
    if (children.Count = 0) then
    begin
        start_x := xmin + BOX_WID;
    end else begin
        // This is where the children will start.
        start_y := start_y + BOX_HGT + BOX_VGAP;

        // Position the children.
        for i := 0 to children.Count - 1 do
        begin
            // Position this child.
            child := children.Items[i];
            child.SetPosition(start_x, start_y);

            // Add a little room before the next child.
            start_x := start_x + BOX_HGAP;
        end;

        // Subtract the gap after the last child.
        start_x := start_x - BOX_HGAP;
    end;

    // Center this node over its children.
    position.Left := (xmin + start_x - BOX_WID) div 2;
    position.Right  := position.Left + BOX_WID;
end;

// Draw this node.
procedure TBigNode.DrawNode(cvs : TCanvas; selected_node : TBigNode);
var
    text_size : TSize;
begin
    // Select appropriate colors depending on whether
    // we need to highlight this node.
    if (selected_node = Self) then
    begin
        cvs.Font.Color := clWhite;
        cvs.Brush.Color := clBlack;
    end else begin
        cvs.Font.Color := clBlack;
        cvs.Brush.Color := clWhite;
    end;

    // Erase the node's position and draw a box.
    cvs.FillRect(position);
    cvs.Rectangle(
        position.Left, position.Top,
        position.Right, position.Bottom);

    // Draw the node's value text.
    text_size := cvs.TextExtent(node_value);
    cvs.TextOut(
        (position.Left + position.Right  - text_size.cx) div 2,
        (position.Top  + position.Bottom - text_size.cy) div 2,
        node_value);
end;

// Draw the subtree rooted at this node.
procedure TBigNode.DrawSubtree(cvs : TCanvas; selected_node : TBigNode);
var
    i     : Integer;
    child : TBigNode;
begin
    // Draw this node.
    DrawNode(cvs, selected_node);

    // Draw the children.
    for i := 0 to children.Count - 1 do
    begin
        child := children.Items[i];

        // Draw the node itself.
        child.DrawSubtree(cvs, selected_node);

        // Draw lines connecting the node and the child.
        cvs.MoveTo(
            position.Left + BOX_WID div 2,
            position.Top + BOX_HGT);
        cvs.LineTo(
            child.position.Left + BOX_WID div 2,
            child.position.Top);
    end;
end;

// Find the descendant of this node that contains
// this point.
//
// Note that this is not a very sophisticated method.
// A quadtree would be much faster. See Ready-to-Run
// Delphi 3.0 Algorithms for more information.
function TBigNode.NodeAtPoint(X, Y : Integer) : TBigNode;
var
    i     : Integer;
    child : TBigNode;
begin
    // Assume we will not find the node.
    Result := nil;

    // See if the point is in this node.
    if ((position.Left <= X) and (X <= position.Right) and
        (position.Top  <= Y) and (Y <= position.Bottom))
    then begin
        // We are the node.
        Result := Self;
    end else begin
        // Recursively check the children.
        for i := 0 to children.Count - 1 do
        begin
            child := children.Items[i];
            Result := child.NodeAtPoint(X, Y);
            if (Result <> nil) then break;
        end;
    end;
end;

// Remove the given node from this subtree. Return True
// if the node was removed.
function TBigNode.RemoveNode(target_node : TBigNode) : Boolean;
var
    i     : Integer;
    child : TBigNode;
begin
    // Assume we will fail to find the target.
    Result := False;

    // Examine each child.
    for i := 0 to children.Count - 1 do
    begin
        child := children.Items[i];

        // See if the child is the target.
        if (child = target_node) then
        begin
            // The child is the target. Remove it.
            children.Delete(i);
            Result := True;
            break;
        end else begin
            // Recursively make the child
            // look for the target.
            Result := child.RemoveNode(target_node);

            // If the child removed the target, we're done.
            if (Result) then break;
        end;
    end;
end;

// Find the node with a particular value.
function TBigNode.FindNodeValue(target_value : String10) : TBigNode;
var
    i     : Integer;
    child : TBigNode;
begin
    // Assume we will not find the target.
    Result := nil;

    // See if this is the node we want.
    if (node_value = target_value) then
    begin
        // It is. Return this node.
        Result := Self;
    end else begin
        // It is not. Make the children search for it.
        for i := 0 to children.Count - 1 do
        begin
            child := children.Items[i];

            // See if this child can find the target.
            Result := child.FindNodeValue(target_value);

            // If the child found it, we are done.
            if (Result <> nil) then break;
        end;
    end;
end;

end.
