{
  Name    :   Transactions.pas

  Purpose :   MTS TAccouts object implementing ITransactions and operating on
               the TRANSACT table.

  Usage   :   This source code is supplied as example code with Delphi Informant
               magazine.
              The author makes no guarantee for its correctness or fitness
               for any purpose and cannot be held responsible for any loss
               or damage that may occur through its use or misuse.
              You may copy, modify or use any part of this source code
               provided the author is credited for his work.

  Author  :   Paul M. Fairhurst, September 1998. paul@c-s-c.demon.co.uk
}
unit Transactions;

interface

uses
  BDEMTS, TransactionsDM, ActiveX, MtsObj, ComObj, DelphiBankServer2_TLB;

type
  TTransactions = class(TMtsAutoObject, ITransactions)
  private
    dmTransactions : TdmTransactions;
  protected
    function AddNew(ClientKey, AcctID: Integer; const TrnRef: WideString;
      TrnAmount: Currency; var strDebug: WideString): Integer; safecall;
    function Delete(ClientKey, TrnID: Integer;
      var strDebug: WideString): Integer; safecall;
    function ListAll(ClientKey: Integer; var varResultSet: OleVariant;
      var strDebug: WideString): Integer; safecall;
    function ListByAcctID(ClientKey, AcctID: Integer;
      var varResultSet: OleVariant; var strDebug: WideString): Integer;
      safecall;
    function ListStatementByAcctID(ClientKey, AcctID: Integer;
      var varResultSet: OleVariant; var strDebug: WideString): Integer;
      safecall;
    { Protected declarations }
    procedure OnActivate; override;
    procedure OnDeactivate; override;
    procedure CheckDbOpen;
  end;

implementation

uses ComServ, ServerUtils, {DMDelphiBank, }SysUtils, DelphiBankCommonDM, Windows;

{ ---------------------------------------------------------------------------- }
procedure TTransactions.OnActivate;
begin
  TServerUtils.WriteToDelphiBankLog('TTransactions - OnActivate entered.');

  try
    { Create our own datamodule. }
    dmTransactions := TdmTransactions.Create(nil);
    if Assigned(dmTransactions) then
      dmTransactions.DbDelphiBank.Open;
  except
    on E: Exception do
      TServerUtils.WriteToDelphiBankLog('TTransactions.OnActivate - ' + E.Message);
  end;
end;

{ ---------------------------------------------------------------------------- }
procedure TTransactions.OnDeactivate;
begin
  TServerUtils.WriteToDelphiBankLog('TTransactions - OnDeactivate entered.');

  try
    if Assigned(dmTransactions) then begin
      dmTransactions.DbDelphiBank.Close;
      dmTransactions.Free;
    end;
    dmTransactions := nil;
  except
  end;
end;

{ ---------------------------------------------------------------------------- }
procedure TTransactions.CheckDbOpen;
begin
  if not Assigned(dmTransactions)then
    raise Exception.Create('Datamodule not instantiated.');
  if not dmTransactions.DbDelphiBank.Connected then
    raise Exception.Create('Could not establish connection to database.');
end;

{ ---------------------------------------------------------------------------- }
function TTransactions.AddNew(ClientKey, AcctID: Integer;
  const TrnRef: WideString; TrnAmount: Currency;
  var strDebug: WideString): Integer;
begin
  Result := cErrResult; { Default. }
  strDebug     := '';

  try
    CheckDbOpen;

    { Set up the query. }
    with dmTransactions do begin
      { Security. }
      if not IsClientKeyValid(ClientKey, strDebug) then Exit;

      { Add the transaction. }
      QryTransactionInsert.ParamByName(cFieldTrnID).Value  := GetTickCount;
      QryTransactionInsert.ParamByName(cFieldAcctID).Value := AcctID;
      QryTransactionInsert.ParamByName(cFieldTrnRef).Value := Trnref;
      QryTransactionInsert.ParamByName(cFieldTrnAmount).Value := TrnAmount;
      QryTransactionInsert.ParamByName(cFieldTrnTimestamp).Value := Now;
      QryTransactionInsert.ExecSQL;
    end;

    SetComplete; { OK. }
    Result := cOKResult;
  except on E: Exception do begin
      SetAbort; { Problem encountered. }
      strDebug := 'TTransactions.AddNew() - ' + E.Message;
      Result := cErrResult;
    end;
  end;

  { Write out to event log. }
  if strDebug <> '' then
    TServerUtils.WriteToDelphiBankLog(strDebug);
end;

{ ---------------------------------------------------------------------------- }
function TTransactions.Delete(ClientKey, TrnID: Integer;
  var strDebug: WideString): Integer;
begin
  Result := cErrResult; { Default. }
  strDebug     := '';

  try
    CheckDbOpen;

    { Set up the query. }
    with dmTransactions do begin
      { Security. }
      if not IsClientKeyValid(ClientKey, strDebug) then Exit;

      { Run the query. }
      QryTransactionDelete.ParamByName(cFieldTrnID).Value := TrnID;
      QryTransactionDelete.ExecSQL;
    end;

    SetComplete; { OK. }
    Result := cOKResult;
  except on E: Exception do begin
      SetAbort; { Problem encountered. }
      strDebug := 'TTransactions.Delete() - ' + E.Message;
      Result := cErrResult;
    end;
  end;

  { Write out to event log. }
  if strDebug <> '' then
    TServerUtils.WriteToDelphiBankLog(strDebug);
end;

{ ---------------------------------------------------------------------------- }
function TTransactions.ListAll(ClientKey: Integer;
  var varResultSet: OleVariant; var strDebug: WideString): Integer;
begin
  Result := cErrResult; { Default. }

  varResultSet := UnAssigned;
  strDebug     := '';

  try
    CheckDbOpen;

    { Set up the query. }
    with dmTransactions do begin
      { Security. }
      if not IsClientKeyValid(ClientKey, strDebug) then Exit;

      { Run the query. }
      QryTransactionReadAll.Open;

      { Package the result. }
      CreateVarArrayFromDataset(varResultSet, QryTransactionReadAll);
      QryTransactionReadAll.Close;
    end;

    SetComplete; { OK. }
    Result := cOKResult;
  except on E: Exception do begin
      SetAbort; { Problem encountered. }
      dmTransactions.QryTransactionReadAll.Close;
      strDebug := 'TTransactions.ListAll() - ' + E.Message;
      Result := cErrResult;
    end;
  end;

  { Write out to event log. }
  if strDebug <> '' then
    TServerUtils.WriteToDelphiBankLog(strDebug);
end;

{ ---------------------------------------------------------------------------- }
function TTransactions.ListByAcctID(ClientKey, AcctID: Integer;
  var varResultSet: OleVariant; var strDebug: WideString): Integer;
begin
  Result := cErrResult; { Default. }

  varResultSet := UnAssigned;
  strDebug     := '';

  try
    CheckDbOpen;

    { Set up the query. }
    with dmTransactions do begin
      { Security. }
      if not IsClientKeyValid(ClientKey, strDebug) then Exit;

      { Run the query. }
      QryTransactionReadByAcctID.ParamByName(cFieldAcctID).Value := AcctID;
      QryTransactionReadByAcctID.Open;

      { Package the result. }
      CreateVarArrayFromDataset(varResultSet, QryTransactionReadByAcctID);
      QryTransactionReadByAcctID.Close;
    end;

    SetComplete; { OK. }
    Result := cOKResult;
  except on E: Exception do begin
      SetAbort; { Problem encountered. }
      dmTransactions.QryTransactionReadByAcctID.Close;
      strDebug := 'TTransactions.ListByAcctID() - ' + E.Message;
      Result := cErrResult;
    end;
  end;

  { Write out to event log. }
  if strDebug <> '' then
    TServerUtils.WriteToDelphiBankLog(strDebug);
end;

{ ---------------------------------------------------------------------------- }
function TTransactions.ListStatementByAcctID(ClientKey, AcctID: Integer;
  var varResultSet: OleVariant; var strDebug: WideString): Integer;
var
  AccountsIntf   : IAccounts;
  ATID, CustID   : Integer;
  AcctMinBalance, AcctBalance : Currency;
  AcctLastStatement    : TDateTime;
begin
  Result := cErrResult; { Default. }

  varResultSet := UnAssigned;
  strDebug     := '';

  try
    CheckDbOpen;

    { Security. }
    if not dmTransactions.IsClientKeyValid(ClientKey, strDebug) then Exit;

    { We need to ask an Account object for its details to get the last
       statement date. }
    try
      OleCheck(ObjectContext.CreateInstance(CLASS_Accounts, IAccounts, AccountsIntf));
    except
      raise Exception.Create('Could not create Accounts object');
    end;

    { Make the call. }
    if AccountsIntf.GetAcctByID(ClientKey, AcctID, ATID, CustID, AcctMinBalance,
                                AcctBalance, AcctLastStatement, strDebug) <> cOKResult then
      raise Exception.Create(strDebug);

    { Set up the query. }
    with dmTransactions do begin
      { Run the query. }
      QryTransactionReadStatementByAcctID.ParamByName(cFieldAcctID).Value := AcctID;
      QryTransactionReadStatementByAcctID.ParamByName(cFieldAcctLastStatement).Value := AcctLastStatement;
      QryTransactionReadStatementByAcctID.Open;

      { Package the result. }
      CreateVarArrayFromDataset(varResultSet, QryTransactionReadStatementByAcctID);
      QryTransactionReadStatementByAcctID.Close;
    end;

    SetComplete; { OK. }
    AccountsIntf := nil;
    Result := cOKResult;
  except on E: Exception do begin
      SetAbort; { Problem encountered. }
      dmTransactions.QryTransactionReadStatementByAcctID.Close;
      AccountsIntf := nil;
      strDebug := 'TTransactions.ListStatementByAcctID() - ' + E.Message;
      Result := cErrResult;
    end;
  end;

  { Write out to event log. }
  if strDebug <> '' then
    TServerUtils.WriteToDelphiBankLog(strDebug);
end;

{ ---------------------------------------------------------------------------- }
initialization
  TAutoObjectFactory.Create(ComServer, TTransactions, Class_Transactions,
    ciMultiInstance, tmApartment);

finalization
  TServerUtils.WriteToDelphiBankLog('Transactions module finalization...');

end.
