unit RepParam;

{
  Ask user for parameters for a report.
  Start up the nominated report.

  Written by Keith Wood.
  Version 1.0 - 9 June 1998.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, DB, DBTables, Menus;

const
  { Constants for parameter values }
  sCustomer = 'Customer';
  sState    = 'State';
  sOrder    = 'Order';
  sDate     = 'Date';
  sDateFrom = 'DateFrom';
  sDateTo   = 'DateTo';
  sTitle    = 'Title';
  sFilter   = 'Filter';
  sSortBy   = 'SortBy';
  sSortDesc = 'SortDesc';

type
  { Possible parameters to be requested from user }
  TReportParameter = (rpCustomer, rpState, rpOrder, rpDate, rpDateRange);
  TReportParameters = set of TReportParameter;

  { The parameters form.
    Each set of parameters resides on its own panel
    which is made visible if required for this report.
    All these panels are aligned alTop. As they are
    made invisible the rest move up automatically.
    The panel at the bottom is always visible and
    is used to set the form height appropriately. }
  TfrmRepParams = class(TForm)
    pnlCustomer: TPanel;
    Label13: TLabel;
    cmbCustomer: TComboBox;
    pnlState: TPanel;
    Label2: TLabel;
    cmbState: TComboBox;
    pnlOrder: TPanel;
    Label12: TLabel;
    cmbOrder: TComboBox;
    pnlDate: TPanel;
    Label7: TLabel;
    edtDate: TEdit;
    btnDate: TBitBtn;
    pnlDateRange: TPanel;
    Label1: TLabel;
    edtDateFrom: TEdit;
    btnDateFrom: TBitBtn;
    Label3: TLabel;
    edtDateTo: TEdit;
    btnDateTo: TBitBtn;
    pnlSortBy: TPanel;
    Label4: TLabel;
    cmbSortBy: TComboBox;
    pnlButtons: TPanel;
    Label6: TLabel;
    radScreen: TRadioButton;
    radPrinter: TRadioButton;
    btnPrinter: TBitBtn;
    lblPrinter: TLabel;
    btnOK: TBitBtn;
    btnCancel: TBitBtn;
    qryCustomer: TQuery;
    qryState: TQuery;
    qryOrder: TQuery;
    dlgPrinterSetup: TPrinterSetupDialog;
    procedure cmbCustomerChange(Sender: TObject);
    procedure btnDateClick(Sender: TObject);
    procedure btnDateFromClick(Sender: TObject);
    procedure btnDateToClick(Sender: TObject);
    procedure radPrintToClick(Sender: TObject);
    procedure btnPrinterClick(Sender: TObject);
    procedure btnOKClick(Sender: TObject);
  private
    iRepIndex: Integer;
    procedure PopupCalendar(edt: TEdit);
    function ConvertDate(sDate: String): TDateTime;
  public
    { Construct a drop-down menu for the reports known to this unit }
    class procedure BuildMenu(mniParent: TMenuItem);
    { Respond to selection of a menu item for a report }
    class procedure ReportMenuClick(Sender: TObject);
    constructor Create(AOwner: TComponent; iReportNum: Integer);
    destructor Destroy; override;
  end;

implementation

{$R *.DFM}

uses
  Printers, RepBase, PopCal;

{ Known reports ---------------------------------------------------------------}

type
  TReportRec =
    record
      { Menu entry under which to appear, format: 'upper level|lower level' }
      MenuEntry: String;
      { Name of QuickReport form }
      Report: String;
      { Title of the report }
      Title: String;
      { Set of user parameters available for entry }
      Params: TReportParameters;
      { Set of parameters that are required - should be subset of above }
      Required: TReportParameters;
      { Possible orderings, format: '<display>=<field(s)>' + sSep + ... }
      SortBy: String;
      { Filter to be applied to where clause of query }
      Filter: String;
    end;

const
  sSep = #13#10;    { Separator for sort by options }
  iNumReports = 7;  { Number of entries in array below }
  { Records containing details of each report - place in order of appearance on menu }
  recReports: array [1..iNumReports] of TReportRec =
    ((MenuEntry: '&Customers|&List'; Report: 'TqrfReport1';
      Title: 'Customer List'; Params: [rpCustomer, rpState]; Required: [];
      SortBy: 'Customer No=CustNo' + sSep + 'Company=Company' + sSep +
        'State=State, Company' + sSep + 'Zip Code=Zip, Company'),
    (MenuEntry: '&Customers|Last &Invoiced'; Report: 'TqrfReport1';
      Title: 'Customer Last Invoiced'; Params: [rpState, rpDate]; Required: [rpDate];
      SortBy: 'Customer No=CustNo' + sSep + 'Company=Company' + sSep +
        'Last Invoiced Date=LastInvoiceDate desc' + sSep +
        'State=State, Company' + sSep + 'Zip Code=Zip, Company'),
    (MenuEntry: '&Orders|&List by Customer'; Report: 'TqrfReport2';
      Title: 'Orders by Customer'; Params: [rpCustomer]; Required: [rpCustomer];
      SortBy: 'Order No=OrderNo' + sSep + 'Sale Date=SaleDate, OrderNo' + sSep +
        'Total Amount=ItemsTotal, OrderNo'),
    (MenuEntry: '&Orders|For &Dates'; Report: 'TqrfReport2';
      Title: 'Orders for Dates'; Params: [rpDateRange]; Required: [rpDateRange];
      SortBy: 'Sale Date=SaleDate, OrderNo' + sSep + 'Customer No=CustNo' + sSep +
        'Order No=OrderNo' + sSep + 'Total Amount=ItemsTotal, OrderNo'),
    (MenuEntry: '&Orders|&Paid'; Report: 'TqrfReport2';
      Title: 'Paid Orders'; Params: [rpState, rpDateRange]; Required: [rpDateRange];
      SortBy: 'Sale Date=SaleDate, OrderNo' + sSep + 'Customer No=CustNo' + sSep +
        'Order No=OrderNo' + sSep + 'Total Amount=ItemsTotal, OrderNo';
      Filter: 'ItemsTotal <= AmountPaid'),
    (MenuEntry: '&Orders|&Unpaid'; Report: 'TqrfReport2';
      Title: 'Unpaid Orders'; Params: [rpState, rpDateRange]; Required: [rpDateRange];
      SortBy: 'Sale Date=SaleDate, OrderNo' + sSep + 'Customer No=CustNo' + sSep +
        'Order No=OrderNo' + sSep + 'Total Amount=ItemsTotal, OrderNo';
      Filter: 'ItemsTotal > AmountPaid'),
    (MenuEntry: '&Orders|&Individual'; Report: 'TqrfReport3';
      Title: 'Individual Order'; Params: [rpCustomer, rpOrder];
      Required: [rpCustomer, rpOrder]));

{ TfrmRepParams ---------------------------------------------------------------}

type
  { Wrapper around a string }
  TString = class(TObject)
  public
    Value: String;
    constructor Create(sValue: String);
  end;

{ Initialise internal string }
constructor TString.Create(sValue: String);
begin
  inherited Create;
  Value := sValue;
end;

{ TfrmRepParams ---------------------------------------------------------------}

{ Construct a menu for the reports known to this unit
  and attach underneath the specified menu item }
class procedure TfrmRepParams.BuildMenu(mniParent: TMenuItem);
var
  i, iPos: Integer;
  mniAddTo, mniItem: TMenuItem;
  sCaption: String;

  { Find upper level menu - creating it if necessary }
  function FindMenuItem(mniCurrent: TMenuItem; sCaption: String): TMenuItem;
  var
    i: Integer;
  begin
    { Check current menu items }
    for i := 0 to mniCurrent.Count - 1 do
      if mniCurrent.Items[i].Caption = sCaption then
      begin
        Result := mniCurrent.Items[i];
        Exit;
      end;
    { Not there - create a new item }
    Result := TMenuItem.Create(mniCurrent);
    try
      Result.Caption := sCaption;
      mniCurrent.Add(Result);
    except
      Result.Free;
      raise;
    end;
  end;

begin
  { Process full array of reports }
  for i := 1 to iNumReports do
    with recReports[i] do
    begin
      { Locate parent menu (if applicable) }
      mniAddTo := mniParent;
      sCaption := MenuEntry;
      iPos := Pos('|', sCaption);
      while iPos > 0 do  { May be several levels to menu }
      begin
        mniAddTo := FindMenuItem(mniAddTo, Copy(sCaption, 1, iPos - 1));
        sCaption := Copy(sCaption, iPos + 1, Length(sCaption) - iPos);
        iPos := Pos('|', sCaption);
      end;
      { Add new menu item }
      mniItem := TMenuItem.Create(mniAddTo);
      try
        with mniItem do
        begin
          Caption := sCaption;
          Tag := i;                    { Index into table here }
          OnClick := ReportMenuClick;  { Return to this unit to process }
        end;
        mniAddTo.Add(mniItem);
      except
        mniItem.Free;
        raise;
      end;
    end;
end;

{ Respond to selection of a menu item for a report }
class procedure TfrmRepParams.ReportMenuClick(Sender: TObject);
var
  iReport: Integer;
begin
  { Tag property is index into array here }
  iReport := TMenuItem(Sender).Tag;
  if not (iReport in [1..iNumReports]) then
    raise Exception.Create('Unknown report tag = ' + IntToStr(iReport));

  { Bring up report parameter screen for nominated report }
  with TfrmRepParams.Create(Application, iReport) do
    try
      ShowModal;
    finally
      Free;
    end;
end;

{ Initialise dialog box based on details from array above and report number }
constructor TfrmRepParams.Create(AOwner: TComponent; iReportNum: Integer);

  { Load customer records for selection }
  procedure LoadCustomers(bVisible: Boolean);
  begin
    pnlCustomer.Visible := bVisible;
    if bVisible then
      with cmbCustomer, qryCustomer do
      begin
        Items.Clear;
        Items.AddObject('', nil);  { Add blank entry for 'all' }
        Open;
        while not EOF do
        begin
          Items.AddObject(FieldByName('Company').AsString,
            Pointer(FieldByName('CustNo').AsInteger));
          Next;
        end;
        Close;
      end;
  end;

  { Load state records for selection }
  procedure LoadStates(bVisible: Boolean);
  begin
    pnlState.Visible := bVisible;
    if bVisible then
      with cmbState, qryState do
      begin
        Items.Clear;
        Items.Add('');  { Add blank entry for 'all' }
        Open;
        while not EOF do
        begin
          Items.Add(FieldByName('State').AsString);
          Next;
        end;
        Close;
      end;
  end;

  { Load order records for selection }
  procedure LoadOrders(bVisible: Boolean);
  begin
    pnlOrder.Visible := bVisible;
    if bVisible then
      cmbCustomerChange(cmbOrder);
  end;

  { Load sorting options }
  procedure LoadSortBy(sSortBy: String);
  var
    slsSort: TStringList;
    i: Integer;
  begin
    slsSort := TStringList.Create;
    try
      { Load sort options from array - already delimited by line feeds
        so Text property loads into separate lines in string list }
      slsSort.Text := sSortBy;
      { Then separate out display and field values }
      for i := 0 to slsSort.Count - 1 do
        cmbSortBy.Items.AddObject(slsSort.Names[i],
          TString.Create(slsSort.Values[slsSort.Names[i]]));
      pnlSortBy.Visible := (cmbSortBy.Items.Count > 1);
      cmbSortBy.ItemIndex := 0;
    finally
      slsSort.Free;
    end;
  end;

begin
  inherited Create(AOwner);
  { Set report index in array of details }
  if not (iReportNum in [1..iNumReports]) then
    raise Exception.Create('Unknown report tag = ' + IntToStr(iReportNum));
  iRepIndex := iReportNum;

  try
    Screen.Cursor := crHourglass;
    { Initialise dialog box }
    with recReports[iRepIndex] do
    begin
      Caption := Title + ' Parameters';
      LoadCustomers(rpCustomer in Params);
      LoadStates(rpState in Params);
      LoadOrders(rpOrder in Params);
      pnlDate.Visible := (rpDate in Params);
      pnlDateRange.Visible := (rpDateRange in Params);
      LoadSortBy(SortBy);
    end;
    cmbOrder.Enabled := not pnlCustomer.Visible;
    ClientHeight := pnlButtons.Top + pnlButtons.Height;
    edtDate.Text := DateTimeToStr(Date);
    edtDateFrom.Text := edtDate.Text;
    edtDateTo.Text := edtDate.Text;
    lblPrinter.Caption := Printer.Printers[Printer.PrinterIndex];
    radPrintToClick(radScreen);
  finally
    Screen.Cursor := crDefault;
  end;
end;

{ Free resources }
destructor TfrmRepParams.Destroy;
var
  i: Integer;
begin
  for i := 0 to cmbSortBy.Items.Count - 1 do
    cmbSortBy.Items.Objects[i].Free;
  inherited Destroy;
end;

{ Convert string to date - default to 0 }
function TfrmRepParams.ConvertDate(sDate: String): TDateTime;
begin
  try
    Result := StrToDate(sDate);
  except
    Result := 0;
  end;
end;

{ Filter orders based on customer }
procedure TfrmRepParams.cmbCustomerChange(Sender: TObject);
var
  iStart, iEnd: Integer;
begin
  if pnlOrder.Visible then
  begin
    if cmbCustomer.ItemIndex > 0 then  { Customer selected }
    begin
      iStart := Integer(cmbCustomer.Items.Objects[cmbCustomer.ItemIndex]);
      iEnd := iStart;
    end
    else                               { All orders }
    begin
      iStart := 0;
      iEnd := MaxInt;
    end;
    with cmbOrder, qryOrder do
    begin
      Items.Clear;
      Items.AddObject('', nil);  { Add blank entry for 'all' }
      ParamByName('CustStart').AsInteger := iStart;
      ParamByName('CustEnd').AsInteger := iEnd;
      Open;
      while not EOF do
      begin
        Items.AddObject(FieldByName('OrderNo').AsString + ' on ' +
          FormatDateTime('MM/DD/YYYY', FieldByName('SaleDate').AsDateTime) +
          ' for ' + Format('%m', [FieldByName('ItemsTotal').AsCurrency]),
          Pointer(FieldByName('OrderNo').AsInteger));
        Next;
      end;
      Close;
    end;
    cmbOrder.ItemIndex := 0;
    cmbOrder.Enabled := not pnlCustomer.Visible or (cmbCustomer.ItemIndex > 0);
  end;
end;

{ Pop up calendar for specified date edit }
procedure TfrmRepParams.PopupCalendar(edt: TEdit);
var
  dDate: TDateTime;
begin
  dDate := TfrmCalendar.Execute(ConvertDate(edt.Text));
  if dDate > 0 then
    edt.Text := DateToStr(dDate);
end;

{ Pop up calendar for date }
procedure TfrmRepParams.btnDateClick(Sender: TObject);
begin
  PopupCalendar(edtDate);
end;

{ Pop up calendar for start date }
procedure TfrmRepParams.btnDateFromClick(Sender: TObject);
begin
  PopupCalendar(edtDateFrom);
end;

{ Pop up calendar for end date }
procedure TfrmRepParams.btnDateToClick(Sender: TObject);
begin
  PopupCalendar(edtDateTo);
end;

{ Dis/enable printer selection }
procedure TfrmRepParams.radPrintToClick(Sender: TObject);
begin
  lblPrinter.Enabled := radPrinter.Checked;
  btnPrinter.Enabled := lblPrinter.Enabled;
end;

{ Allow the user to change printers }
procedure TfrmRepParams.btnPrinterClick(Sender: TObject);
begin
  dlgPrinterSetup.Execute;
  lblPrinter.Caption := Printer.Printers[Printer.PrinterIndex];
end;

{ Submit report with these parameters }
procedure TfrmRepParams.btnOKClick(Sender: TObject);
var
  dDate, dDateFrom, dDateTo: TDateTime;
  slsParams: TStringList;  { List of parameters }

  { Validate entry of required combo box fields }
  procedure ValidateComboEntry(rpParam: TReportParameter; sMessage: String;
    cmb: TComboBox);
  begin
    if (rpParam in recReports[iRepIndex].Required) and (cmb.Text = '') then
    begin
      MessageDlg(sMessage + ' must be selected', mtWarning, [mbOK], 0);
      cmb.SetFocus;
      ModalResult := mrNone;
      Abort;
    end;
  end;

  { Validate all the entered fields }
  procedure ValidateFields;
  begin
    { Validate combo boxes }
    ValidateComboEntry(rpCustomer, 'A customer', cmbCustomer);
    ValidateComboEntry(rpState, 'A state', cmbState);
    ValidateComboEntry(rpOrder, 'An order', cmbOrder);

    { Validate dates }
    dDate := ConvertDate(edtDate.Text);
    if (rpDate in recReports[iRepIndex].Required) and (dDate = 0) then
    begin
      MessageDlg('A date must be selected', mtWarning, [mbOK], 0);
      edtDate.SetFocus;
      ModalResult := mrNone;
      Abort;
    end;

    dDateFrom := ConvertDate(edtDateFrom.Text);
    dDateTo := ConvertDate(edtDateTo.Text);
    if (rpDateRange in recReports[iRepIndex].Required) and
        ((dDateFrom = 0) and (dDateTo = 0)) then
    begin
      MessageDlg('A date range must be selected', mtWarning, [mbOK], 0);
      if edtDateFrom.Text = '' then
        edtDateFrom.SetFocus
      else
        edtDateTo.SetFocus;
      ModalResult := mrNone;
      Abort;
    end;
    if (dDateFrom > 0) and (dDateTo > 0) and (dDateFrom > dDateTo) then
    begin
      MessageDlg('Start date cannot be after end date', mtWarning, [mbOK], 0);
      edtDateFrom.SetFocus;
      ModalResult := mrNone;
      Abort;
    end;
  end;

  { Build parameter list }
  procedure AddUserParameters;
  begin
    { Report title }
    slsParams.Values[sTitle] := recReports[iRepIndex].Title;
    { Sort order, if applicable }
    if cmbSortBy.ItemIndex > -1 then
    begin
      slsParams.Values[sSortBy] :=
        TString(cmbSortBy.Items.Objects[cmbSortBy.ItemIndex]).Value;
      slsParams.Values[sSortDesc] := cmbSortBy.Items[cmbSortBy.ItemIndex];
    end;
    { Report specific filter }
    if recReports[iRepIndex].Filter <> '' then
      slsParams.Values[sFilter] := recReports[iRepIndex].Filter;

    { User entered parameters - combo boxes }
    if (rpCustomer in recReports[iRepIndex].Params) and (cmbCustomer.ItemIndex > 0) then
      slsParams.Values[sCustomer] :=
        IntToStr(Integer(cmbCustomer.Items.Objects[cmbCustomer.ItemIndex]));
    if (rpState in recReports[iRepIndex].Params) and (cmbState.ItemIndex > 0) then
      slsParams.Values[sState] := cmbState.Items[cmbState.ItemIndex];
    if (rpOrder in recReports[iRepIndex].Params) and (cmbOrder.ItemIndex > 0) then
      slsParams.Values[sOrder] :=
        IntToStr(Integer(cmbOrder.Items.Objects[cmbOrder.ItemIndex]));

    { User entered parameters - date }
    if (rpDate in recReports[iRepIndex].Params) and (dDate <> 0) then
      slsParams.Values[sDate] := FormatDateTime('dd-mmm-yy', dDate);

    { User entered parameters - date range }
    if (rpDateRange in recReports[iRepIndex].Params) then
    begin
      if dDateFrom <> 0 then
        slsParams.Values[sDateFrom] := FormatDateTime('dd-mmm-yy', dDateFrom);
      if dDateTo <> 0 then
        slsParams.Values[sDateTo] := FormatDateTime('dd-mmm-yy', dDateTo);
    end;
  end;

begin
  { Force on exit processing }
  btnOK.SetFocus;

  try
    Screen.Cursor := crHourglass;
    btnOK.Enabled := False;
    btnCancel.Enabled := False;
    slsParams := TStringList.Create;

    { Check data entry and prepare parameters }
    ValidateFields;
    AddUserParameters;
    { And run the report }
    with recReports[iRepIndex] do
      TqrfBase.ShowReport(Report, Title, slsParams, radScreen.Checked);
  finally
    slsParams.Free;
    btnOK.Enabled := True;
    btnCancel.Enabled := True;
    Screen.Cursor := crDefault;
  end;
end;

end.
